import json
from hashlib import sha256
from types import SimpleNamespace

import httpx
import pytest

from fbnconfig import cutlabel

TEST_BASE = "https://foo.lusid.com"


@pytest.mark.respx(base_url=TEST_BASE)
class DescribeCutLabelRef:
    client = httpx.Client(base_url=TEST_BASE, event_hooks={"response": [lambda x: x.raise_for_status()]})

    def test_attach_when_present(self, respx_mock):
        # given that the remote definition exists
        respx_mock.get("/api/api/systemconfiguration/cutlabels/cd1").mock(
            return_value=httpx.Response(200, json={})
        )
        client = self.client
        sut = cutlabel.CutLabelRef(
            id="one",
            code="cd1"
        )
        # when we call attach
        sut.attach(client)
        # then a get request was made and no exception raised
        req = respx_mock.calls[0]
        assert req.request.url == f"{TEST_BASE}/api/api/systemconfiguration/cutlabels/cd1"

    def test_attach_when_absent(self, respx_mock):
        # given that the remote definition does not exist
        respx_mock.get("/api/api/systemconfiguration/cutlabels/cd1").mock(
            return_value=httpx.Response(404, json={})
        )
        client = self.client
        sut = cutlabel.CutLabelRef(
            id="one",
            code="cd1"
        )
        # when we call attach
        with pytest.raises(RuntimeError) as ex:
            sut.attach(client)
        # then a get request was made
        req = respx_mock.calls[0]
        assert req.request.url == f"{TEST_BASE}/api/api/systemconfiguration/cutlabels/cd1"
        # and the exception explains
        assert str(ex.value) == "CutLabel cd1 not found"

    def test_attach_when_http_error(self, respx_mock):
        # given that the remote read fails with 500
        respx_mock.get("/api/api/systemconfiguration/cutlabels/cd1").mock(
            return_value=httpx.Response(500, json={})
        )
        client = self.client
        sut = cutlabel.CutLabelRef(
            id="one",
            code="cd1"
        )
        # when we call attach it throws
        with pytest.raises(httpx.HTTPStatusError):
            sut.attach(client)


class DescribeCutLabelResource:
    client = httpx.Client(base_url=TEST_BASE, event_hooks={"response": [lambda x: x.raise_for_status()]})

    def get_hash(self, obj: cutlabel.CutLabelResource) -> str:
        desired = obj.model_dump(mode="json", exclude_none=True, by_alias=True)
        sorted_desired = json.dumps(desired, sort_keys=True)
        return sha256(sorted_desired.encode()).hexdigest()

    def test_read(self, respx_mock):
        # given that the remote definition exists
        respx_mock.get("/api/api/systemconfiguration/cutlabels/cd1").mock(
            return_value=httpx.Response(200, json={})
        )
        client = self.client
        sut = cutlabel.CutLabelResource(
            id="one",
            code="cd1",
            display_name="foo",
            cut_local_time=cutlabel.CutTime(hours=1, minutes=2, seconds=3.4),
            time_zone="UTC"
        )
        # when we call read
        sut.read(client, SimpleNamespace(code="cd1"))
        # then a get request was made and no exception raised
        req = respx_mock.calls[0]
        assert req.request.url == f"{TEST_BASE}/api/api/systemconfiguration/cutlabels/cd1"

    def test_create(self, respx_mock):
        # given that the remote definition does not exist
        respx_mock.post("/api/api/systemconfiguration/cutlabels").mock(
            return_value=httpx.Response(201, json={})
        )
        client = self.client
        sut = cutlabel.CutLabelResource(
            id="one",
            code="cd1",
            display_name="foo",
            cut_local_time=cutlabel.CutTime(hours=1, minutes=2, seconds=3.4),
            time_zone="UTC"
        )
        # when we call create
        new_state = sut.create(client)
        # then a post request was made to create the remote
        post = respx_mock.calls[0]
        assert post.request.url == f"{TEST_BASE}/api/api/systemconfiguration/cutlabels"
        assert json.loads(post.request.content) == {
            "code": "cd1",
            "displayName": "foo",
            "cutLocalTime": {"hours": 1, "minutes": 2, "seconds": 3.4},
            "timeZone": "UTC"
        }
        # and the new state is returned
        assert new_state == {"code": "cd1", "content_hash": self.get_hash(sut)}

    def test_delete(self, respx_mock):
        # given that the remote definition exists
        respx_mock.delete("/api/api/systemconfiguration/cutlabels/cd1").mock(
            return_value=httpx.Response(204, json={})
        )
        client = self.client
        sut = cutlabel.CutLabelResource(
            id="one",
            code="cd1",
            display_name="foo",
            cut_local_time=cutlabel.CutTime(hours=1, minutes=2, seconds=3.4),
            time_zone="UTC"
        )
        # when we call delete
        sut.delete(client, SimpleNamespace(code="cd1", content_hash=self.get_hash(sut)))
        # then a delete request was made
        req = respx_mock.calls[0]
        assert req.request.url == f"{TEST_BASE}/api/api/systemconfiguration/cutlabels/cd1"

    def test_delete_special_chars(self, respx_mock):
        # given that the remote definition exists and has a slash in the code
        respx_mock.delete("/api/api/systemconfiguration/cutlabels/cd%2F1").mock(
            return_value=httpx.Response(204, json={})
        )
        client = self.client
        # when we call delete
        cutlabel.CutLabelResource.delete(client, SimpleNamespace(code="cd/1", content_hash="whatever"))
        # then a delete request was made with the code escaped in the url
        req = respx_mock.calls[0]
        assert req.request.url == f"{TEST_BASE}/api/api/systemconfiguration/cutlabels/cd%2F1"

    def test_update_no_change(self):
        # given the remote state has the same has as the desired state
        client = self.client
        sut = cutlabel.CutLabelResource(
            id="one",
            code="cd1",
            display_name="foo",
            cut_local_time=cutlabel.CutTime(hours=1, minutes=2, seconds=3.4),
            time_zone="UTC"
        )
        remote_state = SimpleNamespace(code="cd1", content_hash=self.get_hash(sut))
        # when we call update with no change
        new_state = sut.update(client, remote_state)
        # then no request is made and the state returned is None
        assert not new_state

    def test_update_changed_hash_only(self, respx_mock):
        # given the remote state has a different hash to the desired state
        respx_mock.put("/api/api/systemconfiguration/cutlabels/cd1").mock(
            return_value=httpx.Response(200, json={})
        )
        client = self.client
        sut = cutlabel.CutLabelResource(
            id="one",
            code="cd1",
            display_name="foo",
            cut_local_time=cutlabel.CutTime(hours=1, minutes=2, seconds=3.4),
            time_zone="UTC"
        )
        remote_state = SimpleNamespace(code="cd1", content_hash="old")
        # when we call update
        new_state = sut.update(client, remote_state)
        # then a put request was made
        req = respx_mock.calls[0]
        assert req.request.url == f"{TEST_BASE}/api/api/systemconfiguration/cutlabels/cd1"
        assert req.request.method == "PUT"
        assert json.loads(req.request.content) == {
            "code": "cd1",
            "displayName": "foo",
            "cutLocalTime": {"hours": 1, "minutes": 2, "seconds": 3.4},
            "timeZone": "UTC"
        }
        # and the new state is returned with a new hash
        assert new_state == {"code": "cd1", "content_hash": self.get_hash(sut)}

    def test_update_change_local_time(self, respx_mock):
        # given the remote state has a different hash to the desired state
        respx_mock.put("/api/api/systemconfiguration/cutlabels/cd1").mock(
            return_value=httpx.Response(200, json={})
        )
        client = self.client
        sut = cutlabel.CutLabelResource(
            id="one",
            code="cd1",
            display_name="foo",
            cut_local_time=cutlabel.CutTime(hours=1, minutes=2, seconds=3.4),
            time_zone="UTC"
        )
        remote_state = SimpleNamespace(code="cd1", content_hash=self.get_hash(sut))
        # and the desired has a different local time
        sut.cut_local_time = cutlabel.CutTime(hours=2, minutes=3, seconds=4.5)
        # when we update
        new_state = sut.update(client, remote_state)
        # then a put request is made
        req = respx_mock.calls[0]
        assert req.request.url == f"{TEST_BASE}/api/api/systemconfiguration/cutlabels/cd1"
        assert req.request.method == "PUT"
        assert json.loads(req.request.content) == {
            "code": "cd1",
            "displayName": "foo",
            "cutLocalTime": {"hours": 2, "minutes": 3, "seconds": 4.5},
            "timeZone": "UTC"
        }
        # and the new state is returned with a new hash
        assert new_state == {"code": "cd1", "content_hash": self.get_hash(sut)}

    def test_update_change_code(self, respx_mock):
        # given the remote state has a different code to the desired state
        respx_mock.delete("/api/api/systemconfiguration/cutlabels/oldcode").mock(
            return_value=httpx.Response(204, json={})
        )
        respx_mock.post("/api/api/systemconfiguration/cutlabels").mock(
            return_value=httpx.Response(201, json={})
        )
        client = self.client
        sut = cutlabel.CutLabelResource(
            id="one",
            code="newcode",  # This is the new desired code
            display_name="foo",
            cut_local_time=cutlabel.CutTime(hours=1, minutes=2, seconds=3.4),
            time_zone="UTC"
        )
        remote_state = SimpleNamespace(code="oldcode", content_hash="whatever")
        # when we call update
        new_state = sut.update(client, remote_state)
        # then the remote is deleted
        delete_req = respx_mock.calls[0]
        assert delete_req.request.url == f"{TEST_BASE}/api/api/systemconfiguration/cutlabels/oldcode"
        assert delete_req.request.method == "DELETE"
        # and a new cutlabel is created
        create_req = respx_mock.calls[1]
        assert create_req.request.url == f"{TEST_BASE}/api/api/systemconfiguration/cutlabels"
        assert create_req.request.method == "POST"
        assert json.loads(create_req.request.content) == {
            "code": "newcode",
            "displayName": "foo",
            "cutLocalTime": {"hours": 1, "minutes": 2, "seconds": 3.4},
            "timeZone": "UTC"
        }
        # and the new state is returned with the new code and hash
        assert new_state == {"code": "newcode", "content_hash": self.get_hash(sut)}

    def test_deps(self):
        # when we call deps
        sut = cutlabel.CutLabelResource(
            id="one",
            code="cd1",
            display_name="foo",
            cut_local_time=cutlabel.CutTime(hours=1, minutes=2, seconds=3.4),
            time_zone="UTC"
        )
        # then it returns an empty list
        assert sut.deps() == []

    def test_dump_undump_with_id_context(self):
        # given a cutlabel resource
        sut = cutlabel.CutLabelResource(
            id="test-cutlabel",
            code="cl1",
            display_name="Test CutLabel",
            description="A test cutlabel for dump/undump",
            cut_local_time=cutlabel.CutTime(hours=17, minutes=30, seconds=0.0),
            time_zone="Europe/London"
        )
        # when we dump it
        dumped = sut.model_dump(
            mode="json",
            by_alias=True,
            exclude_none=True,
            round_trip=True,
            context={"style": "dump"}
        )
        # then the id should be excluded from the dump
        assert "id" not in dumped
        assert dumped == {
            "code": "cl1",
            "displayName": "Test CutLabel",
            "description": "A test cutlabel for dump/undump",
            "cutLocalTime": {"hours": 17, "minutes": 30, "seconds": 0.0},
            "timeZone": "Europe/London"
        }
        # when we undump it with id from context
        result = cutlabel.CutLabelResource.model_validate(
            dumped, context={"style": "dump", "id": "context-cutlabel"}
        )
        # then it should be correctly populated including id from context
        assert result.id == "context-cutlabel"
        assert result.code == "cl1"
        assert result.display_name == "Test CutLabel"
        assert result.description == "A test cutlabel for dump/undump"
        assert result.cut_local_time.hours == 17
        assert result.cut_local_time.minutes == 30
        assert result.cut_local_time.seconds == 0.0
        assert result.time_zone == "Europe/London"



from fbnconfig import Deployment, side_definition, transaction_type


def configure(env):
    deployment_name = getattr(env, "name", "transaction_configuration")

    test_side1 = side_definition.SideResource(
        id="side1",
        side="Side1",
        scope=deployment_name,
        security="Txn:LusidInstrumentId",
        currency="Txn:SettlementCurrency",
        rate="Txn:TradeToPortfolioRate",
        units="Txn:Units",
        amount="Txn:TotalConsideration",
        notional_amount="0",
    )

    test_side2 = side_definition.SideResource(
        id="side2",
        side="Side2",
        scope=deployment_name,
        security="Txn:SettlementCurrency",
        currency="Txn:SettlementCurrency",
        rate="SettledToPortfolioRate",
        units="Txn:TotalConsideration",
        amount="Txn:TotalConsideration",
        notional_amount="0",
    )

    test_buy_transaction_type = transaction_type.TransactionTypeResource(
        id="example",
        scope=deployment_name,
        source="default",
        aliases=[
            transaction_type.TransactionTypeAlias(
                type="Buy",
                description="Something",
                transaction_class="default",
                transaction_roles="LongLonger",
                is_default=False,
            ),
            transaction_type.TransactionTypeAlias(
                type="BY",
                description="Something but different",
                transaction_class="default",
                transaction_roles="LongLonger",
                is_default=False,
            ),
        ],
        movements=[
            transaction_type.TransactionTypeMovement(
                movement_types=transaction_type.MovementType.StockMovement,
                side=test_side1,
                direction=1,
                name="Stock Movement",
                movement_options=[transaction_type.MovementOption.DirectAdjustment],
            )
        ],
        calculations=[
            transaction_type.TransactionTypeCalculation(
                type=transaction_type.CalculationType.TaxAmounts, side=test_side2
            ),
            transaction_type.TransactionTypeCalculation(
                type=transaction_type.CalculationType.NotionalAmount
            ),
        ],
    )

    return Deployment(deployment_name, [test_buy_transaction_type])


if __name__ == "__main__":
    import os

    import click

    import fbnconfig

    @click.command()
    @click.argument("lusid_url", envvar="LUSID_ENV", type=str)
    @click.option("-v", "--vars_file", type=click.File("r"))
    def cli(lusid_url, vars_file):
        host_vars = fbnconfig.load_vars(vars_file)
        d = configure(host_vars)
        fbnconfig.deploy(d, lusid_url, os.environ["FBN_ACCESS_TOKEN"])

    cli()



import json
from hashlib import sha256
from typing import Annotated, Any, Dict, List, Literal, Sequence

import httpx
from pydantic import BaseModel, BeforeValidator, Field, PlainSerializer, computed_field, model_validator

from fbnconfig.coretypes import ResourceId

from . import property
from .resource_abc import CamelAlias, Ref, Resource, register_resource


class ReferenceListRef(CamelAlias, BaseModel, Ref):
    id: str
    scope: str
    code: str

    def attach(self, client):
        try:
            url = f"/api/api/referencelists/{self.scope}/{self.code}"
            response = client.get(url)
            return response.json()
        except httpx.HTTPStatusError as ex:
            if ex.response.status_code == 404:
                return None
            else:
                raise ex


class AddressKeyList(CamelAlias, BaseModel):
    values: list[str]
    reference_list_type: Literal["AddressKeyList"] = Field("AddressKeyList", init=False)


class DecimalList(CamelAlias, BaseModel):
    values: list[float]
    reference_list_type: Literal["DecimalList"] = Field("DecimalList", init=False)


class FundIdList(CamelAlias, BaseModel):
    values: list[ResourceId]
    reference_list_type: Literal["FundIdList"] = Field("FundIdList", init=False)


class PortfolioGroupIdList(CamelAlias, BaseModel):
    values: list[ResourceId]
    reference_list_type: Literal["PortfolioGroupIdList"] = Field("PortfolioGroupIdList", init=False)


class PortfolioIdList(CamelAlias, BaseModel):
    values: list[ResourceId]
    reference_list_type: Literal["PortfolioIdList"] = Field("PortfolioIdList", init=False)


class InstrumentList(CamelAlias, BaseModel):
    values: list[str]
    reference_list_type: Literal["InstrumentList"] = Field("InstrumentList", init=False)


class StringList(CamelAlias, BaseModel):
    values: list[str]
    reference_list_type: Literal["StringList"] = Field("StringList", init=False)


class MetricValue(BaseModel):
    value: float
    unit: str | None


class PropertyValue(CamelAlias, BaseModel):
    label_value: str | None = None  # noqa: N815
    metric_value: MetricValue | None = None  # noqa: N815
    label_set_value: Sequence[str] | None = None  # noqa: N815

    @model_validator(mode="after")
    def validate_one_value_exists(self):
        fields = ["label_value", "metric_value", "label_set_value"]
        s = [field for field in fields if getattr(self, field) is not None]
        if len(s) > 1:
            raise KeyError(f"Cannot set {' and '.join(s)}, only one of {' or '.join(fields)} can be set")
        return self


class PropertyListItem(CamelAlias, BaseModel):
    key: property.PropertyKey
    value: PropertyValue
    effective_from: str | None = None
    effective_until: str | None = None


class PropertyList(CamelAlias, BaseModel):
    values: list[PropertyListItem]
    reference_list_type: Literal["PropertyList"] = Field("PropertyList", init=False)


ReferenceListTypes = Annotated[
    AddressKeyList | DecimalList | FundIdList | PortfolioGroupIdList |
    PortfolioIdList | InstrumentList | StringList | PropertyList,
    Field(discriminator="reference_list_type")
]


@register_resource()
class ReferenceListResource(BaseModel, Resource):
    id: str = Field(exclude=True)
    scope: str
    code: str
    name: str
    description: str | None = None
    tags: list[str] | None = None
    reference_list: ReferenceListTypes

    @computed_field(alias="id")
    def list_id(self) -> dict[str, str]:
        return {"scope": self.scope, "code": self.code}

    @model_validator(mode="before")
    @classmethod
    def des_model(cls, data: Any, info) -> Dict[str, Any]:
        if not isinstance(data, dict):
            return data
        if isinstance(data.get("id"), dict):
            # promote scope and code from id dict
            data = data | data["id"]
        if info.context and info.context.get("id"):
            # Handle id from context (for dump/undump)
            # For undump, the id might be a string that needs to be the resource id
            context_id = info.context.get("id")
            data = data | {"id": context_id}
        return data

    def read(self, client, old_state):
        scope, code = old_state.scope, old_state.code
        url = f"/api/api/referencelists/{scope}/{code}"
        entity = client.get(url).json()
        entity.pop("version", None)
        return entity

    def create(self, client) -> Dict[str, Any]:
        desired = self.model_dump(mode="json", exclude_none=True, by_alias=True,
                                  exclude={"id", "scope", "code"})
        sorted_desired = json.dumps(desired, sort_keys=True)
        content_hash = sha256(sorted_desired.encode()).hexdigest()
        client.request("POST", "/api/api/referencelists", json=desired)
        return {"scope": self.scope, "code": self.code, "content_hash": content_hash}

    def update(self, client, old_state):
        if (self.scope, self.code) != (old_state.scope, old_state.code):
            self.delete(client, old_state)
            return self.create(client)
        desired = self.model_dump(mode="json", exclude_none=True, by_alias=True
                                  , exclude={"id", "scope", "code"})
        sorted_desired = json.dumps(desired, sort_keys=True)
        desired_hash = sha256(sorted_desired.encode()).hexdigest()
        if desired_hash == old_state.content_hash:
            return None
        client.request("POST", "/api/api/referencelists", json=desired)
        return {"scope": self.scope, "code": self.code, "content_hash": desired_hash}

    @staticmethod
    def delete(client, old_state):
        client.request("DELETE", f"/api/api/referencelists/{old_state.scope}/{old_state.code}")

    def deps(self) -> List[Resource | Ref]:
        if isinstance(self.reference_list, PropertyList):
            props = {i.key.id: i.key for i in self.reference_list.values}
            return list(props.values())
        return []


def ser_referencelist_key(value, info):
    print("ser_resource_id", value)
    if info.context and info.context.get("style") == "dump":
        return {"$ref": value.id}
    return {"scope": value.scope, "code": value.code}


def des_referencelist_key(value, info):
    if not isinstance(value, dict):
        return value
    if info.context and info.context.get("$refs"):
        ref = info.context["$refs"][value["$ref"]]
        return ref
    return value


ReferenceListKey = Annotated[
    ReferenceListResource | ReferenceListRef,
    BeforeValidator(des_referencelist_key), PlainSerializer(ser_referencelist_key)
]



import json
import string
from enum import StrEnum
from hashlib import sha256
from typing import List, Sequence

from pydantic import BaseModel, Field, field_serializer, model_validator

from . import property
from .resource_abc import CamelAlias, Ref, Resource, register_resource
from .urlfmt import Urlfmt

url: string.Formatter = Urlfmt("api/api/fundconfigurations")
CREATE_FUNDCONFIGURATION_URL = ("{base}/{scope}")
FUND_CONFIGURATION_URL = ("{base}/{scope}/{code}")


class MetricValue(BaseModel):
    value: float
    unit: str | None


class PropertyValueInner(CamelAlias, BaseModel):
    label_value: str | None
    metric_value: MetricValue | None
    label_set_value: Sequence[str] | None

    @model_validator(mode="after")
    def validate_one_value_exists(self):
        fields = ["label_value", "metric_value", "label_set_value"]
        s = [field for field in fields if getattr(self, field) is not None]
        if len(s) > 1:
            raise KeyError(f"Cannot set {' and '.join(s)}, only one of {' or '.join(fields)} can be set")
        return self


class PropertyValue(CamelAlias, BaseModel):
    key: property.PropertyKey  # Renamed from property_definition
    value: PropertyValueInner
    effective_from: str | None = None
    effective_until: str | None = None

    @classmethod
    def create(cls, key, label_value=None, metric_value=None,
                    label_set_value=None, effective_from=None, effective_until=None):
        """Convenience method to create PropertyValue with the nested value structure"""
        value_inner = PropertyValueInner(
            label_value=label_value,
            metric_value=metric_value,
            label_set_value=label_set_value
        )
        return cls(
            key=key,
            value=value_inner,
            effective_from=effective_from,
            effective_until=effective_until
        )


class AppliesToEnum(StrEnum):
    Undefined = "Undefined"
    PnLBucket = "PnLBucket"
    Fees = "Fees"


class ExternalFeeComponentFilter(CamelAlias, BaseModel):
    filter_id: str
    filter: str
    applies_to: AppliesToEnum


class ComponentFilter(CamelAlias, BaseModel):
    filter_id: str
    filter: str


@register_resource()
class FundConfigurationResource(BaseModel, Resource):
    id: str = Field(exclude=True)
    scope: str
    # request body
    code: str
    display_name: str | None = None
    description: str | None = None
    dealing_filters: List[ComponentFilter]
    pnl_filters: List[ComponentFilter]
    back_out_filters: List[ComponentFilter]
    external_fee_filters: List[ExternalFeeComponentFilter] | None = None
    properties: List[PropertyValue] | None = None

    @field_serializer("properties", mode="plain")
    def serialize_properties(self, properties, _info):
        if properties is None:
            return None
        return {
            f"FundConfiguration/{pv.key.scope}/{pv.key.code}": pv for pv in properties
        }

    @model_validator(mode="before")
    @classmethod
    def des_model(cls, data, info):
        if not isinstance(data, dict):
            return data

        if info.context and info.context.get("id"):
            data = data | {"id": info.context.get("id")}
        return data

    def read(self, client, old_state):
        scope, code = old_state.scope, old_state.code
        formatted_url = url.format(FUND_CONFIGURATION_URL, scope=scope, code=code)
        return client.get(formatted_url).json()

    def create(self, client):
        desired = self.model_dump(mode="json", exclude_none=True, by_alias=True)
        sorted_desired = json.dumps(desired, sort_keys=True)
        content_hash = sha256(sorted_desired.encode("utf-8")).hexdigest()

        scope = desired["scope"]
        code = desired["code"]

        formatted_url = url.format(CREATE_FUNDCONFIGURATION_URL, scope=scope)
        client.post(formatted_url, json=desired)

        return {"scope": scope, "code": code, "content_hash": content_hash}

    def update(self, client, old_state):
        # Currently only some fields can be updated with FundConfiguration
        # So we must delete and recreate a fund configuration if other fields change
        if (self.scope, self.code) != (old_state.scope, old_state.code):
            self.delete(client, old_state)
            return self.create(client)
        desired = self.model_dump(mode="json", exclude_none=True, by_alias=True)
        sorted_desired = json.dumps(desired, sort_keys=True)
        desired_hash = sha256(sorted_desired.encode("utf-8")).hexdigest()
        if desired_hash == old_state.content_hash:
            return None
        self.delete(client, old_state)
        return self.create(client)

    @staticmethod
    def delete(client, old_state):
        scope, code = old_state.scope, old_state.code
        formatted_url = url.format(FUND_CONFIGURATION_URL, scope=scope, code=code)
        client.delete(formatted_url)

    def deps(self) -> Sequence[Resource | Ref]:
        property_keys = []
        if self.properties is None:
            return property_keys

        for property_def in self.properties:
            if property_def.key is not None:
                property_keys.append(property_def.key)

        return property_keys



import os

import fbnconfig
import tests.integration.application as application

if os.getenv("FBN_ACCESS_TOKEN") is None or os.getenv("LUSID_ENV") is None:
    raise (
        RuntimeError("Both FBN_ACCESS_TOKEN and LUSID_ENV variables need to be set to run these tests")
    )

lusid_env = os.environ["LUSID_ENV"]
token = os.environ["FBN_ACCESS_TOKEN"]
host_vars = {}


def test_teardown():
    deployment_name = application.configure({}).id
    fbnconfig.deploy(fbnconfig.Deployment(deployment_name, []), lusid_env, token)
    client = fbnconfig.create_client(lusid_env, token)
    matches = get_applications_by_client_id(client, "robTest-app-client")
    assert len(matches) == 0


def test_create():
    fbnconfig.deploy(application.configure(host_vars), lusid_env, token)
    client = fbnconfig.create_client(lusid_env, token)
    matches = get_applications_by_client_id(client, "robTest-app-client")
    assert len(matches) == 1
    app = matches[0]
    assert app["displayName"] == "robTest Application"
    assert app["type"] == "Native"


def test_update():
    fbnconfig.deploy(application.configure(host_vars), lusid_env, token)
    client = fbnconfig.create_client(lusid_env, token)
    matches = get_applications_by_client_id(client, "robTest-app-client")
    assert len(matches) == 1
    app = matches[0]
    assert app["displayName"] == "robTest Application"
    assert app["type"] == "Native"


def get_applications_by_client_id(client, client_id):
    get = client.request("get", "/identity/api/applications")
    get.raise_for_status()
    applications = get.json()
    return [app for app in applications if app["clientId"] == client_id]



import os
from types import SimpleNamespace

from pytest import fixture

import fbnconfig
from fbnconfig.corporate_action import CorporateActionSourceResource
from tests.integration.generate_test_name import gen

if os.getenv("FBN_ACCESS_TOKEN") is None or os.getenv("LUSID_ENV") is None:
    raise (
        RuntimeError("Both FBN_ACCESS_TOKEN and LUSID_ENV variables need to be set to run these tests")
    )

lusid_env = os.environ["LUSID_ENV"]
token = os.environ["FBN_ACCESS_TOKEN"]
host_vars = {}
client = fbnconfig.create_client(lusid_env, token)


@fixture()
def base_resources(setup_deployment):
    deployment_name = setup_deployment.name
    # Create a corporate action source
    corporate_action_source = CorporateActionSourceResource(
        id="test_corporate_action",
        scope=deployment_name,
        code="test-ca-source-1",
        display_name="Test Corporate Action Source",
        description="A test corporate action source for integration testing",
        instrument_scopes=["robtest"]
    )
    return [corporate_action_source]


@fixture(scope="module")
def setup_deployment():
    deployment_name = gen("corporate_action")
    print(f"\nRunning for deployment {deployment_name}...")
    yield SimpleNamespace(name=deployment_name)
    # Clean up corporate action sources after tests
    try:
        client.delete(f"/api/api/corporateactionsources/{deployment_name}/test-ca-source-1")
    except Exception:
        pass  # Ignore cleanup errors


def test_create(setup_deployment, base_resources):
    deployment = fbnconfig.Deployment(setup_deployment.name, base_resources)
    fbnconfig.deploy(deployment, lusid_env, token)
    # Verify corporate action source was created
    filter_query = f"id.scope eq '{setup_deployment.name}' and id.code eq 'test-ca-source-1'"
    response = client.get("/api/api/corporateactionsources", params={"filter": filter_query}).json()
    values = response["values"]
    assert len(values) == 1
    source = values[0]
    assert source["id"]["scope"] == setup_deployment.name
    assert source["id"]["code"] == "test-ca-source-1"
    assert source["displayName"] == "Test Corporate Action Source"
    assert source["description"] == "A test corporate action source for integration testing"
    assert source["instrumentScopes"] == ["robtest"]


def test_nochange(setup_deployment, base_resources):
    # given we have deployed the base case
    deployment = fbnconfig.Deployment(setup_deployment.name, base_resources)
    fbnconfig.deploy(deployment, lusid_env, token)
    # when we apply it again
    update = fbnconfig.deploy(deployment, lusid_env, token)
    # then there are no changes
    ca_changes = [a.change for a in update if a.type == "CorporateActionSourceResource"]
    assert ca_changes == ["nochange"]


def test_teardown(setup_deployment, base_resources):
    deployment_name = setup_deployment.name
    # given we have deployed the base case
    deployment = fbnconfig.Deployment(deployment_name, base_resources)
    fbnconfig.deploy(deployment, lusid_env, token)
    # when we remove all the resources
    empty = fbnconfig.Deployment(deployment_name, [])
    update = fbnconfig.deploy(empty, lusid_env, token)
    # then there are no changes
    ca_changes = [a.change for a in update if a.type == "CorporateActionSourceResource"]
    assert ca_changes == ["remove"]


def test_update(setup_deployment, base_resources):
    deployment_name = setup_deployment.name
    # Given we have deployed the base case
    initial = fbnconfig.Deployment(deployment_name, base_resources)
    fbnconfig.deploy(initial, lusid_env, token)
    # when we update a resource
    updated_resources = [
        CorporateActionSourceResource(
            id="test_corporate_action",
            scope=deployment_name,
            code="test-ca-source-1",
            display_name="Updated Corporate Action Source",  # Changed display name
            description="An updated corporate action source",  # Changed description
            instrument_scopes=["robupdated"]  # Added scope
        )
    ]
    # and deploy it
    updated_deployment = fbnconfig.Deployment(deployment_name, updated_resources)
    update = fbnconfig.deploy(updated_deployment, lusid_env, token)
    # then we expect the resource to change
    ca_changes = [a.change for a in update if a.type == "CorporateActionSourceResource"]
    assert ca_changes == ["update"]
    # and it has the new values
    filter_query = f"id.scope eq '{deployment_name}' and id.code eq 'test-ca-source-1'"
    updated_response = client.get("/api/api/corporateactionsources",
                                  params={"filter": filter_query}).json()["values"]
    assert len(updated_response) == 1
    source = updated_response[0]
    assert source["displayName"] == "Updated Corporate Action Source"
    assert source["description"] == "An updated corporate action source"
    assert source["instrumentScopes"] == ["robupdated"]



import time

import httpx


# Throw an exception whenever the response code is 4xx or 5xx
def response_hook(response: httpx.Response) -> None:
    if response.is_error:
        response.read()
        response.raise_for_status()


class RetryTransport(httpx.BaseTransport):
    def __init__(self, **kwargs):
        self._wrapper = httpx.HTTPTransport(**kwargs)

    def handle_request(self, request):
        response = self._wrapper.handle_request(request)
        if response.status_code == 429:
            retry_header = response.headers.get("retry-after", None)
            delay = int(retry_header) if retry_header else 0.5
            time.sleep(delay)
            return self.handle_request(request)
        return response

    def close(self):
        self._wrapper.close()


def create_client(lusid_url: str, token: str) -> httpx.Client:
    if token is None or lusid_url is None:
        raise (RuntimeError("Both FBN_ACCESS_TOKEN and LUSID_ENV variables need to be set"))

    return httpx.Client(
        base_url=lusid_url,
        timeout=60,
        headers={
            "authorization": "Bearer " + token,
            "accept": "application/json",
            "content-type": "application/json",
            "X-LUSID-Application": "fbnconfig",
        },
        event_hooks={"response": [response_hook]},
        transport=RetryTransport(),
    )



import os
import textwrap
from unittest.mock import Mock, patch

import httpx
import pytest
from click.testing import CliRunner

from fbnconfig import main


class DescribeRun:
    @patch.dict(os.environ, {"FBN_ACCESS_TOKEN": "xxxxx"})
    @pytest.mark.respx(base_url="https://foo.lusid.com")
    def test_handles_http_exceptions(self, respx_mock):
        # given a config script that does not deploy any resources
        # (it will only call the log endpoint)
        script = textwrap.dedent("""\
            from fbnconfig import Deployment
            def configure(env):
                return Deployment('test1', [])
        """)
        runner = CliRunner()
        # and the log endpoint which returns 500
        respx_mock.get("/api/api/customentities/~deployment").mock(
            return_value=httpx.Response(500, json={})
        )
        with runner.isolated_filesystem():
            with open("script.py", "w") as f:
                f.write(script)
            # when we run the cli with the script
            result = runner.invoke(
                main.cli, ["run", "-e", "https://foo.lusid.com", "-t", "xxxxyyy", "script.py"]
            )
            # then the exit code is failure
            assert result.exit_code == 1, result.output
            # and the 500 is reported to the console
            assert "The server responded with 500" in result.output

    @patch.dict(os.environ, {"FBN_ACCESS_TOKEN": "xxxxx"})
    @pytest.mark.respx(base_url="https://foo.lusid.com")
    def test_handles_invalid_token_http_exceptions(self, respx_mock):
        # given a config script that does not deploy any resources
        # (it will only call the log endpoint)
        script = textwrap.dedent("""\
                from fbnconfig import Deployment
                def configure(env):
                    return Deployment('test1', [])
            """)
        runner = CliRunner()
        # and the log endpoint which returns 401 due to invalid token
        respx_mock.get("/api/api/customentities/~deployment").mock(
            return_value=httpx.Response(401, json={})
        )
        with runner.isolated_filesystem():
            with open("script.py", "w") as f:
                f.write(script)
            # when we run the cli with the script
            result = runner.invoke(
                main.cli, ["run", "-e", "https://foo.lusid.com", "-t", "xxxxyyy", "script.py"]
            )
            # then the exit code is failure
            assert result.exit_code == 1, result.output
            # and the 401 is reported to the console
            assert "The server responded with 401" in result.output
            assert (
                "Unauthorized. This may indicate that "
                "the provided access token is invalid or expired."
            ) in result.output

    @patch.dict(os.environ, {"FBN_ACCESS_TOKEN": "xxxxx"})
    @pytest.mark.respx(base_url="https://foo.lusid.com")
    def test_handles_http_exceptions_with_404_detail(self, respx_mock):
        # given a config script that does not deploy any resources
        # (it will only call the log endpoint)
        script = textwrap.dedent("""\
               from fbnconfig import Deployment
               def configure(env):
                   return Deployment('test1', [])
           """)
        runner = CliRunner()
        # and the log endpoint which returns 404
        respx_mock.get("/api/api/customentities/~deployment").mock(
            return_value=httpx.Response(404, json={"detail": "Some message"})
        )
        with runner.isolated_filesystem():
            with open("script.py", "w") as f:
                f.write(script)
            # when we run the cli with the script
            result = runner.invoke(
                main.cli, ["run", "-e", "https://foo.lusid.com", "-t", "xxxxyyy", "script.py"]
            )
            # then the exit code is failure
            assert result.exit_code == 1, result.output
            # and the detail message is reported to the console
            assert "Some message" in result.output

    @patch.dict(os.environ, {"FBN_ACCESS_TOKEN": "xxxxx"})
    @pytest.mark.respx(base_url="https://foo.lusid.com")
    def test_handles_access_denied_exceptions(self, respx_mock):
        # given a config script that does not deploy any resources
        # (it will only call the log endpoint)
        script = textwrap.dedent("""\
                    from fbnconfig import Deployment
                    def configure(env):
                        return Deployment('test1', [])
                """)
        runner = CliRunner()
        # and the log endpoint which returns 401 due to invalid token
        respx_mock.get("/api/api/customentities/~deployment").mock(
            return_value=httpx.Response(
                403,
                json={
                    "name": "AccessDenied",
                    "errorDetails": [],
                    "status": 403,
                    "detail": "Access to perform the requested action cannot "
                    "be granted due to insufficient privileges. "
                    "Please contact your organisation's "
                    "administrator.",
                    "instance": "https://foo.lusid.com" "/app/insights/logs/0AB123:001A",
                },
            )
        )
        with runner.isolated_filesystem():
            with open("script.py", "w") as f:
                f.write(script)
            # when we run the cli with the script
            result = runner.invoke(
                main.cli, ["run", "-e", "https://foo.lusid.com", "-t", "xxxxyyy", "script.py"]
            )
            # then the exit code is failure
            assert result.exit_code == 1, result.output
            # and the 403 is reported to the console
            assert "The server responded with 403" in result.output
            assert (
                "Access to perform the requested action cannot be "
                "granted due to insufficient privileges"
            ) in result.output

    @patch.dict(os.environ, {"FBN_ACCESS_TOKEN": "xxxxx"})
    @patch("fbnconfig.log.list_resources_for_deployment", return_value=[])
    def test_loads_module(self, mock_log):
        script = textwrap.dedent("""\
            from fbnconfig import Deployment
            def configure(env):
                return Deployment('test1', [])
        """)
        runner = CliRunner()
        with runner.isolated_filesystem():
            with open("script.py", "w") as f:
                f.write(script)
            result = runner.invoke(
                main.cli, ["run", "-e", "https://foo.lusid.com", "-t", "xxxxyyy", "script.py"]
            )
            assert result.exit_code == 0, result.output
            assert len(mock_log.mock_calls) == 1

    @patch.dict(os.environ, {"FBN_ACCESS_TOKEN": "xxxxx"})
    @patch("fbnconfig.log.list_resources_for_deployment", return_value=[])
    def test_loads_module_subfolder(self, mock_log):
        script = textwrap.dedent("""\
            from fbnconfig import Deployment
            def configure(env):
                return Deployment('test1', [])
        """)
        runner = CliRunner()
        with runner.isolated_filesystem():
            os.mkdir("subfolder")
            with open("subfolder/script.py", "w") as f:
                f.write(script)
            result = runner.invoke(
                main.cli, ["run", "-e", "https://foo.lusid.com", "-t", "xxxxyyy", "subfolder/script.py"]
            )
            assert result.exit_code == 0, result.output
            assert len(mock_log.mock_calls) == 1

    @patch.dict(os.environ, {"FBN_ACCESS_TOKEN": "xxxxx"})
    @patch("fbnconfig.log.list_resources_for_deployment", return_value=[])
    def test_error_module_syntax_error(self, mock_log):
        # given a config with a syntax error
        script = textwrap.dedent("""\
            from fbnconfig import NOT_EXISTS
            def configure(env):
                return Deployment('test1', [])
        """)
        runner = CliRunner()
        with runner.isolated_filesystem():
            with open("script.py", "w") as f:
                f.write(script)
            # when we try to run it
            result = runner.invoke(
                main.cli, ["run", "-e", "https://foo.lusid.com", "-t", "xxxxyyy", "script.py"]
            )
            # it fails
            assert result.exit_code == 1, result.output
            # and prints an explantion
            assert "Failed importing" in result.output

    @patch.dict(os.environ, {"FBN_ACCESS_TOKEN": "xxxxx"})
    @patch("fbnconfig.log.list_resources_for_deployment", return_value=[])
    def test_error_module_no_configure_func(self, mock_log):
        # given a config without a configure function
        script = textwrap.dedent("""\
            from fbnconfig import Deployment
            def not_the_configure(env):
                return Deployment('test1', [])
        """)
        runner = CliRunner()
        with runner.isolated_filesystem():
            with open("script.py", "w") as f:
                f.write(script)
            # when we try to run it
            result = runner.invoke(
                main.cli, ["run", "-e", "https://foo.lusid.com", "-t", "xxxxyyy", "script.py"]
            )
            # it fails
            assert result.exit_code == 1, result.output
            # and prints the reason
            assert "No configure function" in result.output

    @patch.dict(os.environ, {"FBN_ACCESS_TOKEN": "xxxxx"})
    @patch("fbnconfig.log.list_resources_for_deployment", return_value=[])
    def test_loads_module_with_import(self, mock_log):
        # given script imports subscript from the same folder
        script = textwrap.dedent("""\
            from fbnconfig import Deployment
            import subscript
            def configure(env):
                grant = subscript.foo()
                return Deployment('test1', [])
        """)
        subscript = textwrap.dedent("""\
            from fbnconfig import access
            def foo():
                return access.Grant.ALLOW
        """)
        runner = CliRunner()
        with runner.isolated_filesystem():
            with open("script.py", "w") as f:
                f.write(script)
            with open("subscript.py", "w") as f:
                f.write(subscript)
            # when we run the cli
            result = runner.invoke(
                main.cli, ["run", "-e", "https://foo.lusid.com", "-t", "xxxxyyy", "script.py"]
            )
            # then it passes showing the import worked
            assert result.exit_code == 0, result.output
            assert len(mock_log.mock_calls) == 1

    @patch.dict(os.environ, {"FBN_ACCESS_TOKEN": "xxxxx"})
    @patch("fbnconfig.log.list_resources_for_deployment", return_value=[])
    def test_loads_module_with_folder_import(self, mock_log):
        # given script imports subscript from another folder
        script = textwrap.dedent("""\
            from fbnconfig import Deployment
            from folder import subscript
            def configure(env):
                grant = subscript.foo()
                return Deployment('test1', [])
        """)
        subscript = textwrap.dedent("""\
            from fbnconfig import access
            def foo():
                return access.Grant.ALLOW
        """)
        runner = CliRunner()
        with runner.isolated_filesystem():
            os.mkdir("folder")
            with open("script.py", "w") as f:
                f.write(script)
            with open("folder/subscript.py", "w") as f:
                f.write(subscript)
            # when we run the cli
            result = runner.invoke(
                main.cli, ["run", "-e", "https://foo.lusid.com", "-t", "xxxxyyy", "script.py"]
            )
            # the import works
            assert result.exit_code == 0, result.output
            assert len(mock_log.mock_calls) == 1


class DescribeSetup:
    @patch("fbnconfig.log.setup")
    def test_with_token_option(self, mock_setup):
        runner = CliRunner()
        result = runner.invoke(main.cli, ["setup", "-e", "https://foo.lusid.com", "-t", "xxxxyyy"])
        assert result.exit_code == 0
        assert len(mock_setup.mock_calls) == 1

    @patch.dict(os.environ, {"FBN_ACCESS_TOKEN": "xxxxx"})
    @patch("fbnconfig.log.setup")
    def test_with_token_env(self, mock_setup):
        runner = CliRunner()
        result = runner.invoke(main.cli, ["setup", "-e", "https://foo.lusid.com"])
        assert result.exit_code == 0
        assert len(mock_setup.mock_calls) == 1

    @patch.dict(os.environ, {"NOT_THE_TOKEN": "xxxxx"}, clear=True)
    @patch("fbnconfig.log.setup")
    def test_without_token(self, mock_setup):
        runner = CliRunner()
        result = runner.invoke(main.cli, ["setup", "-e", "https://foo.lusid.com"])
        assert result.exit_code == 2, result.output
        assert "Missing option '-t'" in result.output
        assert len(mock_setup.mock_calls) == 0

    @patch("fbnconfig.log.setup")
    def test_with_env_option(self, mock_setup):
        runner = CliRunner()
        result = runner.invoke(main.cli, ["setup", "-e", "https://foo.lusid.com", "-t", "xxxxyyy"])
        assert result.exit_code == 0
        assert len(mock_setup.mock_calls) == 1

    @patch.dict(os.environ, {"LUSID_ENV": "https://foo.lusid.com"})
    @patch("fbnconfig.log.setup")
    def test_with_env_env(self, mock_setup):
        runner = CliRunner()
        result = runner.invoke(main.cli, ["setup", "-t", "xxxxyyy"])
        assert result.exit_code == 0
        assert len(mock_setup.mock_calls) == 1

    @patch.dict(os.environ, {"NOT_THE_ENV": "https://foo.lusid.com"}, clear=True)
    @patch("fbnconfig.log.setup")
    def test_without_env(self, mock_setup):
        runner = CliRunner()
        result = runner.invoke(main.cli, ["setup", "-t", "xxxxyyy"])
        assert result.exit_code == 2, result.output
        assert "Missing option '-e'" in result.output
        assert len(mock_setup.mock_calls) == 0

    @pytest.mark.respx(base_url="https://foo.lusid.com")
    def test_handles_http_exceptions(self, respx_mock):
        respx_mock.post("/api/api/propertydefinitions").mock(return_value=httpx.Response(400, json={}))
        runner = CliRunner()
        result = runner.invoke(main.cli, ["setup", "-e", "https://foo.lusid.com", "-t", "x"])
        assert result.exit_code == 1
        # then the exit code is failure
        assert result.exit_code == 1, result.output
        # and the 500 is reported to the console
        assert "The server responded with 400" in result.output


def test_help():
    runner = CliRunner()
    result = runner.invoke(main.cli, ["--help"])
    assert result.exit_code == 0
    assert "Usage" in result.output
    assert "setup" in result.output
    assert "log" in result.output


def test_version():
    runner = CliRunner()
    result = runner.invoke(main.cli, ["--version"])
    assert result.exit_code == 0
    assert "0.0.1" in result.output


class DescribeLogList:
    @patch.dict(
        os.environ, {"FBN_ACCESS_TOKEN": "xxxxx", "LUSID_ENV": "https://foo.lusid.com"}, clear=True
    )
    @patch("fbnconfig.log.list_resources_for_deployment")
    def test_with_token(self, mock_list_resources_for_deployment):
        runner = CliRunner()
        result = runner.invoke(main.cli, ["log", "list", "test"])
        assert result.exit_code == 0
        assert len(mock_list_resources_for_deployment.mock_calls) == 2

    @patch.dict(os.environ, {"NOT_THE_TOKEN": "xxxxx", "LUSID_ENV": "https://foo.lusid.com"}, clear=True)
    @patch("fbnconfig.log.list_deployments")
    def test_without_token(self, mock_list_deployments):
        runner = CliRunner()
        result = runner.invoke(main.cli, ["log", "list"])
        assert result.exit_code == 2, result.output
        assert "Missing option '-t'" in result.output
        assert len(mock_list_deployments.mock_calls) == 0

    @pytest.mark.respx(base_url="https://foo.lusid.com")
    def test_handles_http_exceptions(self, respx_mock):
        respx_mock.get("/api/api/customentities/~deployment").mock(
            return_value=httpx.Response(400, json={})
        )
        runner = CliRunner()
        result = runner.invoke(main.cli, ["log", "list", "-e", "https://foo.lusid.com", "-t", "xxtoken"])
        assert result.exit_code == 1
        # then the exit code is failure
        assert result.exit_code == 1, result.output
        # and the 500 is reported to the console
        assert "The server responded with 400" in result.output


@pytest.mark.respx(base_url="https://foo.lusid.com")
class DescribeLogGet:
    @patch.dict(os.environ, {"FBN_ACCESS_TOKEN": "xx", "LUSID_ENV": "https://foo.lusid.com"}, clear=True)
    @patch("fbnconfig.log.get_resource")
    def test_with_token(self, mock_get_resource):
        runner = CliRunner()
        result = runner.invoke(main.cli, ["log", "get", "test", "test"])
        assert result.exit_code == 0
        assert len(mock_get_resource.mock_calls) == 2

    def test_handles_http_exceptions(self, respx_mock):
        respx_mock.get("/api/api/customentities/~deployment").mock(
            return_value=httpx.Response(400, json={})
        )
        runner = CliRunner()
        result = runner.invoke(
            main.cli,
            ["log", "get", "-e", "https://foo.lusid.com", "-t", "xxtoken", "deployment", "resourceId"],
        )
        # then the exit code is failure
        assert result.exit_code == 1, result.output
        # and the 500 is reported to the console
        assert "The server responded with 400" in result.output


@patch.dict(os.environ, {"FBN_ACCESS_TOKEN": "xxxxx", "LUSID_ENV": "https://foo.lusid.com"}, clear=True)
class DescribeDependencies:
    @patch("fbnconfig.log.get_dependencies_map")
    def test_with_some_dependencies(self, mock_get_deps: Mock):
        mock_get_deps.return_value = {}, {"a": ["b"], "c": ["d"]}
        runner = CliRunner()
        result = runner.invoke(main.cli, ["log", "deps", "deploy", "resource"])
        assert result.exit_code == 0
        mock_get_deps.assert_called_once()


@patch.dict(os.environ, {"FBN_ACCESS_TOKEN": "xxxxx", "LUSID_ENV": "https://foo.lusid.com"}, clear=True)
class DescribeDependants:
    @patch("fbnconfig.log.get_dependents_map")
    def test_with_some_dependencies(self, mock_get_uses_map: Mock):
        mock_get_uses_map.return_value = {}, {"a": ["b", "c"]}
        runner = CliRunner()
        result = runner.invoke(main.cli, ["log", "uses", "deploy", "resource"])
        assert result.exit_code == 0
        mock_get_uses_map.assert_called_once()


@patch.dict(os.environ, {"FBN_ACCESS_TOKEN": "xxxxx", "LUSID_ENV": "https://foo.lusid.com"}, clear=True)
class DescribeRemove:
    @patch("fbnconfig.log.get_dependents_map")
    @patch("fbnconfig.log.get_resource")
    def test_given_resource_does_not_exist(
        self, mock_get_resources: Mock, mock_get_dependents_map: Mock
    ):
        mock_get_resources.return_value = []
        runner = CliRunner()
        result = runner.invoke(main.cli, ["log", "rm", "mydeployment", "someresource"])
        assert result.exit_code == 2, result.output
        mock_get_resources.assert_called_once()
        mock_get_dependents_map.assert_not_called()

    @patch("fbnconfig.log.remove")
    @patch("fbnconfig.log.get_dependents_map")
    @patch("fbnconfig.log.get_resource")
    def test_given_resource_exists_with_uses_and_no_force(
            self, mock_get_resources: Mock, mock_get_dependents_map: Mock, mock_remove: Mock
    ):
        # given resource1 has a dependency on somedeps
        mock_get_resources.return_value = ["someresource"]
        mock_get_dependents_map.return_value = {}, {"someresource": ["somedeps"]}
        # when we try to remove it without a force option
        runner = CliRunner()
        result = runner.invoke(main.cli, ["log", "rm", "mydeployment", "someresource"])
        # then we exit failure, tell the user why
        assert result.exit_code == 2, result.output
        mock_get_dependents_map.assert_called_once()
        assert "Cannot remove entry for 'someresource' as ['somedeps'] depends on it" in result.output
        # and remove does not get called
        mock_remove.assert_not_called()

    @patch("fbnconfig.log.get_resource")
    @patch("fbnconfig.log.remove")
    @patch("fbnconfig.log.get_dependents_map")
    def test_given_resource_exists_with_uses_and_force_then_rm(
        self, mock_get_dependents_map: Mock, mock_remove: Mock, mock_get_resources: Mock
    ):
        mock_get_resources.return_value = ["resource_1"]
        mock_get_dependents_map.return_value = {}, {"someresource": ["somedeps"]}
        runner = CliRunner()
        result = runner.invoke(main.cli, ["log", "rm", "mydeployment", "someresource", "--force"])
        assert result.exit_code == 0
        mock_get_dependents_map.assert_called_once()
        mock_remove.assert_called_once()

    @patch("fbnconfig.log.get_resource")
    @patch("fbnconfig.log.remove")
    @patch("fbnconfig.log.get_dependents_map")
    def test_given_resource_exists_with_uses_and_force_and_recursive_then_rm(
        self, mock_get_dependents_map: Mock, mock_remove: Mock, mock_get_resources: Mock
    ):
        mock_get_resources.return_value = ["resource_1"]
        mock_get_dependents_map.return_value = {}, {"someresource": ["somedeps", "somedeps"]}
        runner = CliRunner()
        result = runner.invoke(
            main.cli, ["log", "rm", "mydeployment", "someresource", "--force", "--recursive"]
        )
        assert result.exit_code == 0
        mock_get_dependents_map.assert_called_once()
        assert mock_remove.call_count == 3



from __future__ import annotations

import datetime as dt
import json
from enum import StrEnum
from hashlib import sha256
from typing import Annotated, Any, Dict, List, Literal, Optional, Self, Sequence, Union

import httpx
from pydantic import BaseModel, BeforeValidator, Field, PlainSerializer, model_validator

from fbnconfig.property import PropertyKey

from .resource_abc import CamelAlias, Ref, Resource, register_resource


class MetricValue(CamelAlias, BaseModel):
    value: float
    unit: Optional[str] = None


def des_isodate(value: str | dt.datetime) -> dt.datetime:
    """Deserialize ISO 8601 date string to datetime object."""
    if isinstance(value, dt.datetime):
        return value.astimezone(tz=dt.timezone.utc)
    return dt.datetime.fromisoformat(value)


def ser_isodate(value: dt.datetime) -> str:
    """Serialize datetime object to ISO 8601 string."""
    return value.isoformat()


IsoDateTime = Annotated[
    dt.datetime | dt.date,
    BeforeValidator(des_isodate), PlainSerializer(ser_isodate)
]


class PropertyValue(CamelAlias, BaseModel):
    property_key: PropertyKey = Field(serialization_alias="propertyKey")
    label_value: Optional[str] = Field(default=None, serialization_alias="labelValue")
    metric_value: Optional[MetricValue] = Field(default=None, serialization_alias="metricValue")
    label_set_value: Optional[List[str]] = Field(default=None, serialization_alias="labelSetValue")
    effective_from: IsoDateTime | None = None
    effective_until: IsoDateTime | None = None

    @model_validator(mode="before")
    @classmethod
    def des_model(cls, data, info):
        if not isinstance(data, dict):
            return data
        style = info.context.get("style", "api") if info.context else "api"
        if style == "api" and data.get("value", None):
            return data | data["value"]
        return data

    @model_validator(mode="after")
    def validate_one_value_exists(self) -> Self:
        fields = ["label_value", "metric_value", "label_set_value"]
        s = [field for field in fields if getattr(self, field) is not None]
        if len(s) > 1:
            raise KeyError(f"Cannot set {' and '.join(s)}, only one of {' or '.join(fields)} can be set")

        return self


def serialize_properties_as_dict(values: List[PropertyValue], info):
    style = info.context.get("style", "api") if info.context else "api"
    if style == "dump":
        return values
    property_dict = {}
    for prop in values:
        value = {}
        if prop.label_value:
            value["labelValue"] = prop.label_value
        elif prop.metric_value:
            value["metricValue"] = prop.metric_value
        elif prop.label_set_value:
            value["labelSetValue"] = prop.label_set_value
        key = f"{prop.property_key.domain.value}/{prop.property_key.scope}/{prop.property_key.code}"
        property_dict[key] = {"key": key, "value": value}

        if prop.effective_from is not None:
            property_dict[key]["effectiveFrom"] = prop.effective_from

        if prop.effective_until is not None:
            property_dict[key]["effectiveUntil"] = prop.effective_until

    return property_dict


def des_properties_dict(data, info) -> Sequence[PropertyValue] | Dict | None:
    if not isinstance(data, dict) or data is None:
        return data
    style = info.context.get("style", "api") if info.context else "api"
    if style == "api" and isinstance(data, dict):
        data = [value | {"propertyKey": value["key"]} for value in data.values()]
    return data


PropertyValueDict = Annotated[
    Sequence[PropertyValue],
    BeforeValidator(des_properties_dict), PlainSerializer(serialize_properties_as_dict)
]


@register_resource()
class ChartOfAccountsRef(BaseModel, Ref):
    """Reference an chart of account

    Example
    -------
    >>> from fbnconfig.fund_accounting import ChartOfAccountsRef
    >>> chart_of_account = ChartOfAccountsRef(
    >>> id="chart_example",
    >>> code="code_example",
    >>> scope="scope_example")
    """

    id: str = Field(exclude=True, init=True)
    code: str
    scope: str

    def attach(self, client):
        try:
            client.request("get", f"/api/api/chartofaccounts/{self.scope}/{self.code}")
        except httpx.HTTPStatusError as ex:
            if ex.response.status_code == 404:
                raise RuntimeError(f"Chart of Accounts {self.scope}/{self.code} not found")
            else:
                raise ex


@register_resource()
class ChartOfAccountsResource(CamelAlias, BaseModel, Resource):
    """Define a Chart of Account in LUSID"""

    id: str = Field(exclude=True)
    scope: str
    code: str
    display_name: str | None = None
    description: str | None = None
    properties: PropertyValueDict | None = None

    @model_validator(mode="before")
    @classmethod
    def des_model(cls, data, info) -> Dict[str, Any]:
        if not isinstance(data, dict):
            return data
        # promote scope and code to top for api style
        if "id" in data and isinstance(data["id"], dict):
            data = data | data["id"]
            data.pop("id", None)
        # Handle id from context (for dump/undump)
        if info.context and info.context.get("id"):
            data = data | {"id": info.context.get("id")}

        return data

    def __get_content_hash__(self) -> str:
        dump = self.model_dump(mode="json", exclude_none=True, by_alias=True)
        return sha256(json.dumps(dump, sort_keys=True).encode()).hexdigest()

    def read(self, client: httpx.Client, old_state) -> None | Dict[str, Any]:
        # Get the property keys to get all properties existing in the chart of accounts
        scope = old_state.scope
        code = old_state.code
        properties = client.get(f"/api/api/chartofaccounts/{scope}/{code}/properties").json()

        params = []
        if "properties" in properties:
            property_keys_list = []
            for key in properties["properties"]:
                property_keys_list.append(key)

            params = [("propertyKeys", prop_key) for prop_key in property_keys_list]

        response = client.get(f"/api/api/chartofaccounts/{scope}/{code}", params=params).json()

        # Remove unnecessary fields
        response.pop("href", None)
        response.pop("links", None)
        return response

    def create(self, client: httpx.Client) -> Optional[Dict[str, Any]]:
        desired = self.model_dump(mode="json", exclude_none=True, by_alias=True, exclude={"scope"})
        response = client.post(f"/api/api/chartofaccounts/{self.scope}", json=desired).json()
        remote_version = response["version"]["asAtVersionNumber"]

        # Remove unnecessary fields
        response.pop("href", None)
        response.pop("links", None)

        return {
            "scope": self.scope,
            "code": self.code,
            "source_version": self.__get_content_hash__(),
            "remote_version": remote_version,
        }

    def update(self, client: httpx.Client, old_state) -> Union[None, Dict[str, Any]]:
        if (self.scope, self.code) != (old_state.scope, old_state.code):
            self.delete(client, old_state)
            return self.create(client)

        source_hash = self.__get_content_hash__()
        remote = self.read(client, old_state)

        remote_hash = None
        if remote is not None:
            remote_hash = remote["version"]["asAtVersionNumber"]

        if remote_hash == old_state.remote_version and source_hash == old_state.source_version:
            return None

        self.delete(client, old_state)
        return self.create(client)

    @staticmethod
    def delete(client: httpx.Client, old_state) -> None:
        client.delete(f"/api/api/chartofaccounts/{old_state.scope}/{old_state.code}")

    def deps(self):
        deps = []
        if self.properties is None:
            return deps

        for perp_prop in self.properties:
            deps.append(perp_prop.property_key)

        return deps


def ser_chart_of_account_key(value, info):
    if info.context and info.context.get("style") == "dump":
        return {"$ref": value.id}
    return {"scope": value.scope, "code": value.code}


def des_chart_of_account_key(value, info):
    if not isinstance(value, dict):
        return value
    if info.context and info.context.get("$refs"):
        ref = info.context["$refs"][value["$ref"]]
        return ref
    return value


ChartOfAccountsKey = Annotated[
    ChartOfAccountsRef | ChartOfAccountsResource,
    BeforeValidator(des_chart_of_account_key), PlainSerializer(ser_chart_of_account_key)
]


class AccountStatus(StrEnum):
    ACTIVE = "Active"
    INACTIVE = "Inactive"
    DELETED = "Deleted"


class AccountType(StrEnum):
    ASSET = "Asset"
    LIABILITIES = "Liabilities"
    INCOME = "Income"
    EXPENSE = "Expense"
    CAPITAL = "Capital"
    REVENUE = "Revenue"


@register_resource()
class AccountRef(CamelAlias, BaseModel, Ref):
    """Reference an account

    Example
    -------
    >>> from fbnconfig.fund_accounting import ChartOfAccountsRef, AccountRef
    >>> chart_of_account = ChartOfAccountsRef(
    >>> id="chart_example",
    >>> code="code_example",
    >>> scope="scope_example")

    >>> account_ref = AccountRef(
    >>> id="account_example",
    >>> account_code="code_example",
    >>> chart_of_accounts=chart_of_account)
    """

    id: str = Field(exclude=True, init=True)
    account_code: str = Field(exclude=True, init=True)
    chart_of_accounts: ChartOfAccountsKey

    def attach(self, client):
        scope = self.chart_of_accounts.scope
        code = self.chart_of_accounts.code

        request_url = f"/api/api/chartofaccounts/{scope}/{code}/accounts/{self.account_code}"
        try:
            client.request("get", request_url)
        except httpx.HTTPStatusError as ex:
            if ex.response.status_code == 404:
                raise RuntimeError(f"Account {self.account_code} not found")
            else:
                raise ex


@register_resource()
class AccountResource(BaseModel, Resource):
    """Define an Account for Chart of Account in LUSID"""

    id: str = Field(exclude=True)
    chart_of_accounts: ChartOfAccountsKey
    code: str
    description: str | None = None
    type: AccountType
    status: AccountStatus
    control: Literal["System", "Manual"] | None = "Manual"
    properties: PropertyValueDict | None = None

    @model_validator(mode="before")
    @classmethod
    def des_model(cls, data, info) -> Dict[str, Any]:
        if not isinstance(data, dict):
            return data
        # Handle id from context (for dump/undump)
        if info.context and info.context.get("id"):
            data = data | {"id": info.context.get("id")}
        return data

    def __get_content_hash__(self) -> str:
        dump = self.model_dump(mode="json", exclude_none=True, by_alias=True)
        return sha256(json.dumps(dump, sort_keys=True).encode()).hexdigest()

    def read(self, client: httpx.Client, old_state) -> None | Dict[str, Any]:
        scope = old_state.scope
        code = old_state.code
        account_code = old_state.account_code

        property_get = f"/api/api/chartofaccounts/{scope}/{code}/accounts/{account_code}/properties"
        properties = client.get(property_get).json()

        params = []
        if "properties" in properties:
            property_keys_list = []
            for key in properties["properties"]:
                property_keys_list.append(key)

            params = [("propertyKeys", prop_key) for prop_key in property_keys_list]

        request_url = f"/api/api/chartofaccounts/{scope}/{code}/accounts/{account_code}"
        return client.get(request_url, params=params).json()

    def create(self, client: httpx.Client) -> Optional[Dict[str, Any]]:
        desired = self.model_dump(
            mode="json",
            exclude_none=True,
            by_alias=True,
            exclude={"chart_of_accounts"})

        desired_wrap = [desired]
        scope = self.chart_of_accounts.scope
        code = self.chart_of_accounts.code

        request_url = f"/api/api/chartofaccounts/{scope}/{code}/accounts"
        response = client.post(request_url, json=desired_wrap).json()

        account_to_dump = response["accounts"][0]
        return {
            "account_code": self.code,
            "scope": scope,
            "code": code,
            "source_version": self.__get_content_hash__(),
            "remote_version": sha256(json.dumps(account_to_dump, sort_keys=True).encode()).hexdigest(),
        }

    def update(self, client: httpx.Client, old_state) -> Union[None, Dict[str, Any]]:
        scope = self.chart_of_accounts.scope
        code = self.chart_of_accounts.code

        if old_state.account_code != self.code:
            self.delete(client, old_state)
            return self.create(client)

        if old_state.code != code:
            raise (RuntimeError("Cannot change the code on an account"))

        if old_state.scope != scope:
            raise (RuntimeError("Cannot change the scope on an accounts"))

        source_hash = self.__get_content_hash__()
        remote = self.read(client, old_state)
        remote_hash = sha256(json.dumps(remote, sort_keys=True).encode()).hexdigest()

        if remote_hash == old_state.remote_version and source_hash == old_state.source_version:
            return None

        return self.create(client)

    @staticmethod
    def delete(client: httpx.Client, old_state) -> None:
        scope = old_state.scope
        code = old_state.code
        account = [old_state.account_code]
        client.post(f"/api/api/chartofaccounts/{scope}/{code}/accounts/$delete", json=account)

    def deps(self):
        deps = []
        deps.append(self.chart_of_accounts)
        deps += self.chart_of_accounts.deps()

        if self.properties is None:
            return deps

        for perp_prop in self.properties:
            deps.append(perp_prop.property_key)

        return deps



import json
from hashlib import sha256
from types import SimpleNamespace

import httpx
import pytest

from fbnconfig import access

TEST_BASE = "https://foo.lusid.com"


def response_hook(response: httpx.Response):
    response.read()
    response.raise_for_status()


@pytest.mark.respx(base_url=TEST_BASE)
class DescribePolicyRef:
    client = httpx.Client(base_url=TEST_BASE, event_hooks={"response": [response_hook]})

    @staticmethod
    def test_scope_default():
        # when scope is omitted
        sut = access.PolicyRef(id="one", code="cd")
        # then it uses "default"
        assert sut.scope == "default"

    def test_attach_when_present(self, respx_mock):
        # given that the remote definition exists
        respx_mock.get("/access/api/policies/pol1?scope=sc1").mock(
            return_value=httpx.Response(200, json={})
        )
        client = self.client
        sut = access.PolicyRef(id="one", scope="sc1", code="pol1")
        # when we call attach
        sut.attach(client)
        # then a get request was made and no exception raised

    def test_attach_when_missing(self, respx_mock):
        # given the remote does not exist
        respx_mock.get("/access/api/policies/pol1?scope=sc1").mock(
            return_value=httpx.Response(404, json={})
        )
        client = self.client
        sut = access.PolicyRef(id="one", scope="sc1", code="pol1")
        # when we call attach an exception is raised
        with pytest.raises(RuntimeError) as ex:
            sut.attach(client)
        assert "Policy sc1/pol1 not found" in str(ex.value)

    def test_attach_when_http_error(self, respx_mock):
        # given the server returns an error
        respx_mock.get("/access/api/policies/pol1?scope=sc1").mock(
            return_value=httpx.Response(500, json={})
        )
        client = self.client
        sut = access.PolicyRef(id="one", scope="sc1", code="pol1")
        # when we call attach an exception is raised
        with pytest.raises(httpx.HTTPError):
            sut.attach(client)


@pytest.mark.respx(base_url=TEST_BASE)
class DescribePolicy:
    client = httpx.Client(base_url=TEST_BASE, event_hooks={"response": [response_hook]})

    @pytest.fixture
    def stock_policy_resource(self):
        # uses new style selector definition: List[Selector]
        return access.PolicyResource(
            id="policy1",
            code="a_policy",
            description="policy description",
            applications=["lusid"],
            grant=access.Grant.ALLOW,
            when=access.WhenSpec(activate="2024-08-31T18:00:00.0000000+00:00"),
            selectors=[
                access.IdSelector(
                    name="banjo",
                    description="whatever",
                    identifier={"scope": "w", "code": "y"},
                    actions=[access.ActionId(scope="actscope", activity="execute", entity="Feature")],
                )
            ],
        )

    def test_create(self, respx_mock):
        respx_mock.post("/access/api/policies").mock(return_value=httpx.Response(200, json={}))
        client = self.client
        # given no policies exist yet
        # when we create one
        sut = access.PolicyResource(
            id="policy1",
            code="a_policy",
            description="policy description",
            applications=["lusid"],
            grant=access.Grant.ALLOW,
            when=access.WhenSpec(activate="2024-08-31T18:00:00.0000000+00:00"),
            selectors=[
                access.IdSelector(
                    name="banjo",
                    description="whatever",
                    identifier={"scope": "w", "code": "y"},
                    actions=[access.ActionId(scope="actscope", activity="execute", entity="Feature")],
                )
            ],
        )
        state = sut.create(client)
        # then the state is returned
        assert state["id"] == "policy1"
        assert state["code"] == "a_policy"
        # and a create request was sent
        request = respx_mock.calls.last.request
        assert request.method == "POST"
        assert request.url.path == "/access/api/policies"
        assert json.loads(request.content) == {
            "applications": ["lusid"],
            "code": "a_policy",
            "description": "policy description",
            "grant": "Allow",
            "selectors": [
                {
                    "idSelectorDefinition": {
                        "actions": [{"activity": "execute", "entity": "Feature", "scope": "actscope"}],
                        "description": "whatever",
                        "identifier": {"code": "y", "scope": "w"},
                        "name": "banjo",
                    }
                }
            ],
            "when": {"activate": "2024-08-31T18:00:00.0000000+00:00"},
        }

    def test_parse_api_format(self):
        # given an api format as returned by /access/api/policy
        remote = {
            "applications": ["lusid"],
            "id": {
                "code": "a_policy",
                "scope": "a_scope",
            },
            "description": "policy description",
            "grant": "Allow",
            "selectors": [
                {
                    "idSelectorDefinition": {
                        "actions": [{"activity": "execute", "entity": "Feature", "scope": "actscope"}],
                        "description": "whatever",
                        "identifier": {"code": "y", "scope": "w"},
                        "name": "banjo",
                    }
                }
            ],
            "when": {"activate": "2024-08-31T18:00:00.0000000+00:00"},
        }
        # when we push it through the deserialization
        converted = access.PolicyResource.model_validate(remote, context={"id": "res123"})
        # then
        assert converted.id == "res123"  # the resource id is from the context
        assert converted.applications == ["lusid"]
        assert converted.scope == "a_scope"  # scope and code get lifted to the top level
        assert converted.code == "a_policy"
        assert type(converted.selectors[0]).__name__ == "IdSelector"  # selector type inferred
        assert converted.selectors[0].type_name == "idSelectorDefinition"
        assert converted.when.activate == "2024-08-31T18:00:00.0000000+00:00"

    def test_dump_policy_resource(self):
        # given a policy resource
        sut = access.PolicyResource(
            id="dump-policy",
            code="DumpPolicy",
            description="A test policy for dumping",
            applications=["lusid"],
            grant=access.Grant.ALLOW,
            when=access.WhenSpec(activate="2024-08-31T18:00:00.0000000+00:00"),
            selectors=[
                access.IdSelector(
                    name="test-selector",
                    description="Test ID selector",
                    identifier={"scope": "test-scope", "code": "test-code"},
                    actions=[access.ActionId(scope="test", activity="execute", entity="Feature")],
                )
            ],
        )
        # when we dump it
        result = sut.model_dump(
            mode="json", by_alias=True, round_trip=True, exclude_none=True, context={"style": "dump"}
        )
        # then all fields are included with proper aliases
        expected = {
            "code": "DumpPolicy",
            "description": "A test policy for dumping",
            "applications": ["lusid"],
            "grant": "Allow",
            "selectors": [
                {
                    "typeName": "idSelectorDefinition",
                    "identifier": {"scope": "test-scope", "code": "test-code"},
                    "actions": [{"scope": "test", "activity": "execute", "entity": "Feature"}],
                    "name": "test-selector",
                    "description": "Test ID selector",
                }
            ],
            "when": {"activate": "2024-08-31T18:00:00.0000000+00:00"},
        }
        assert result == expected

    def test_undump_policy_resource(self):
        # given dumped policy data
        data = {
            "code": "UndumpPolicy",
            "description": "A test policy for undumping",
            "applications": ["lusid"],
            "grant": "Allow",
            "selectors": [
                {
                    "typeName": "idSelectorDefinition",
                    "identifier": {"scope": "undump-scope", "code": "undump-code"},
                    "actions": [{"scope": "undump", "activity": "read", "entity": "Portfolio"}],
                    "name": "undump-selector",
                    "description": "Undump ID selector",
                }
            ],
            "when": {"activate": "2024-12-01T00:00:00.0000000+00:00"},
        }
        # when we undump it
        result = access.PolicyResource.model_validate(data, context={
            "style": "dump", "id": "undump-policy"
        })
        # then the resource Id is extracted from the context
        assert result.id == "undump-policy"
        assert result.code == "UndumpPolicy"
        # and the selector has the right type
        assert len(result.selectors) == 1
        assert isinstance(result.selectors[0], access.IdSelector)
        assert result.selectors[0].type_name == "idSelectorDefinition"
        assert result.when.activate == "2024-12-01T00:00:00.0000000+00:00"

    def test_create_matchall_selector(self, respx_mock):
        respx_mock.post("/access/api/policies").mock(return_value=httpx.Response(200, json={}))
        client = self.client
        # given no policies exist yet
        # when we create one with a matchall selector
        sut = access.PolicyResource(
            id="policy1",
            code="a_policy",
            description="policy description",
            applications=["lusid"],
            grant=access.Grant.ALLOW,
            when=access.WhenSpec(activate="2024-08-31T18:00:00.0000000+00:00"),
            selectors=[
                access.MatchAllSelector(
                    name="banjo",
                    description="whatever",
                    actions=[access.ActionId(scope="actscope", activity="execute", entity="Feature")],
                )
            ],
        )
        state = sut.create(client)
        # then the state is returned
        assert state["id"] == "policy1"
        assert state["code"] == "a_policy"
        # and a create request was sent
        request = respx_mock.calls.last.request
        assert request.method == "POST"
        assert request.url.path == "/access/api/policies"
        assert json.loads(request.content) == {
            "applications": ["lusid"],
            "code": "a_policy",
            "description": "policy description",
            "grant": "Allow",
            "selectors": [
                {
                    "matchAllSelectorDefinition": {
                        "actions": [{"activity": "execute", "entity": "Feature", "scope": "actscope"}],
                        "description": "whatever",
                        "name": "banjo",
                    }
                }
            ],
            "when": {"activate": "2024-08-31T18:00:00.0000000+00:00"},
        }

    def test_create_metadata_selector(self, respx_mock):
        respx_mock.post("/access/api/policies").mock(return_value=httpx.Response(200, json={}))
        client = self.client
        # given no policies exist yet
        # when we create one with a metadata selector
        sut = access.PolicyResource(
            id="policy1",
            code="a_policy",
            description="policy description",
            applications=["lusid"],
            grant=access.Grant.ALLOW,
            when=access.WhenSpec(activate="2024-08-31T18:00:00.0000000+00:00"),
            selectors=[
                access.MetadataSelector(
                    name="banjo",
                    description="whatever",
                    actions=[access.ActionId(scope="actscope", activity="execute", entity="Feature")],
                    expressions=[
                        access.MetadataExpression(
                            metadata_key="foo", operator="Equals", text_value="bingo"
                        )
                    ],
                )
            ],
        )
        state = sut.create(client)
        # then the state is returned
        assert state["id"] == "policy1"
        assert state["code"] == "a_policy"
        # and a create request was sent
        request = respx_mock.calls.last.request
        assert request.method == "POST"
        assert request.url.path == "/access/api/policies"
        assert json.loads(request.content) == {
            "applications": ["lusid"],
            "code": "a_policy",
            "description": "policy description",
            "grant": "Allow",
            "selectors": [
                {
                    "metadataSelectorDefinition": {
                        "actions": [{"activity": "execute", "entity": "Feature", "scope": "actscope"}],
                        "description": "whatever",
                        "name": "banjo",
                        "expressions": [
                            {"metadataKey": "foo", "operator": "Equals", "textValue": "bingo"}
                        ],
                    }
                }
            ],
            "when": {"activate": "2024-08-31T18:00:00.0000000+00:00"},
        }

    def test_create_policy_selector(self, respx_mock):
        respx_mock.post("/access/api/policies").mock(return_value=httpx.Response(200, json={}))
        client = self.client
        # given no policies exist yet
        # when we create one with a policy selector and a nested matchall
        inner_selector = access.MatchAllSelector(
            name="banjo",
            description="whatever",
            actions=[access.ActionId(scope="actscope", activity="execute", entity="Feature")],
        )
        sut = access.PolicyResource(
            id="policy1",
            code="a_policy",
            description="policy description",
            applications=["lusid"],
            grant=access.Grant.ALLOW,
            when=access.WhenSpec(activate="2024-08-31T18:00:00.0000000+00:00"),
            selectors=[
                access.PolicySelector(
                    name="banjo",
                    description="whatever",
                    actions=[access.ActionId(scope="actscope", activity="execute", entity="Feature")],
                    identity_restriction={"id": "value"},
                    restriction_selectors=[inner_selector],
                )
            ],
        )
        state = sut.create(client)
        # then the state is returned
        assert state["id"] == "policy1"
        assert state["code"] == "a_policy"
        # and a create request was sent including the inner selector
        request = respx_mock.calls.last.request
        assert request.method == "POST"
        assert request.url.path == "/access/api/policies"
        assert json.loads(request.content) == {
            "applications": ["lusid"],
            "code": "a_policy",
            "description": "policy description",
            "grant": "Allow",
            "selectors": [
                {
                    "policySelectorDefinition": {
                        "actions": [{"activity": "execute", "entity": "Feature", "scope": "actscope"}],
                        "description": "whatever",
                        "name": "banjo",
                        "identityRestriction": {"id": "value"},
                        "restrictionSelectors": [
                            {
                                "matchAllSelectorDefinition": {
                                    "actions": [
                                        {"activity": "execute", "entity": "Feature", "scope": "actscope"}
                                    ],
                                    "description": "whatever",
                                    "name": "banjo",
                                }
                            }
                        ],
                    }
                }
            ],
            "when": {"activate": "2024-08-31T18:00:00.0000000+00:00"},
        }

    @pytest.fixture
    def read_policy_response(self):
        return {
            "applications": ["lusid"],
            "id": {
                "code": "a_policy_code",
                "scope": "default",
            },  # get returns an ID where post takes a code only
            "description": "policy description",
            "grant": "Allow",
            "selectors": [
                {
                    "idSelectorDefinition": {
                        "actions": [{"activity": "execute", "entity": "Feature", "scope": "actscope"}],
                        "description": "whatever",
                        "identifier": {"code": "y", "scope": "w"},
                        "name": "banjo",
                    }
                }
            ],
            "when": {
                "activate": "2024-08-31T18:00:00.0000000+00:00",
                "deactivate": "99999-12-31T18:00:00.0000000+00:00",
            },
            "links": {},
        }

    def test_update_no_changes(self, respx_mock, read_policy_response):
        # given the policy exists
        respx_mock.get("/access/api/policies/a_policy_code").mock(
            side_effect=[httpx.Response(200, json=read_policy_response)]
        )
        # and the desired is the same as the existing
        sut = access.PolicyResource(
            id="policy1",
            code="a_policy_code",
            description="policy description",
            applications=["lusid"],
            grant=access.Grant.ALLOW,
            when=access.WhenSpec(activate="2024-08-31T18:00:00.0000000+00:00"),
            selectors=[
                access.IdSelector(
                    name="banjo",
                    description="whatever",
                    identifier={"scope": "w", "code": "y"},
                    actions=[access.ActionId(scope="actscope", activity="execute", entity="Feature")],
                )
            ],
        )
        # when we update
        log = sut.update(self.client, SimpleNamespace(code="a_policy_code", id="x"))
        # then there is no change and the log remains as is
        assert log is None
        # and no put request was made

    def test_update_modify_description(self, respx_mock, read_policy_response):
        # given the policy exists with a description = "policy description"
        respx_mock.get("/access/api/policies/a_policy_code").mock(
            side_effect=[httpx.Response(200, json=read_policy_response)]
        )
        respx_mock.put("/access/api/policies/a_policy_code").mock(
            side_effect=[httpx.Response(200, json={})]
        )
        # and the desired has a different description
        sut = access.PolicyResource(
            id="policy1",
            code="a_policy_code",
            description="a different policy description",
            applications=["lusid"],
            grant=access.Grant.ALLOW,
            when=access.WhenSpec(activate="2024-08-31T18:00:00.0000000+00:00"),
            selectors=[
                access.IdSelector(
                    name="banjo",
                    description="whatever",
                    identifier={"scope": "w", "code": "y"},
                    actions=[access.ActionId(scope="actscope", activity="execute", entity="Feature")],
                )
            ],
        )
        # when we update
        log = sut.update(self.client, SimpleNamespace(code="a_policy_code", id="x"))
        # the log is returned (but unchanged)
        assert log == {"code": "a_policy_code", "id": "policy1"}
        # and a put request was made to update the policy
        request = respx_mock.calls.last.request
        assert request.method == "PUT"
        assert request.url.path == "/access/api/policies/a_policy_code"
        assert json.loads(request.content) == {
            "applications": ["lusid"],
            "description": "a different policy description",
            "grant": "Allow",
            "selectors": [
                {
                    "idSelectorDefinition": {
                        "actions": [{"activity": "execute", "entity": "Feature", "scope": "actscope"}],
                        "description": "whatever",
                        "identifier": {"code": "y", "scope": "w"},
                        "name": "banjo",
                    }
                }
            ],
            "when": {"activate": "2024-08-31T18:00:00.0000000+00:00"},
        }

    def test_deps(self, stock_policy_resource):
        # policy doesn't have any deps
        sut = stock_policy_resource
        assert sut.deps() == []

    def test_delete(self, respx_mock):
        respx_mock.delete("/access/api/policies/a_policy_code").mock(
            side_effect=[httpx.Response(200, json={})]
        )
        # given an existing policy
        # when we delete it
        old_state = SimpleNamespace(id="policy1", code="a_policy_code")
        access.PolicyResource.delete(self.client, old_state)
        # then a delete request is made
        request = respx_mock.calls.last.request
        assert request.method == "DELETE"
        assert request.url.path == "/access/api/policies/a_policy_code"


@pytest.mark.respx(base_url=TEST_BASE)
class DescribeRoleRef:
    client = httpx.Client(base_url=TEST_BASE, event_hooks={"response": [response_hook]})

    def test_attach_when_present(self, respx_mock):
        # given that the remote definition exists
        respx_mock.get("/access/api/roles/cd1?scope=sc1").mock(return_value=httpx.Response(200, json={}))
        client = self.client
        sut = access.RoleRef(id="one", scope="sc1", code="cd1")
        # when we call attach
        sut.attach(client)
        # then a get request was made and no exception raised

    def test_attach_when_missing(self, respx_mock):
        # given the remote does not exist
        respx_mock.get("/access/api/roles/cd1?scope=default").mock(
            return_value=httpx.Response(404, json={})
        )
        client = self.client
        sut = access.RoleRef(id="one", code="cd1")
        # when we call attach an exception is raised
        with pytest.raises(RuntimeError) as ex:
            sut.attach(client)
        assert "Role default/cd1 not found" in str(ex.value)

    def test_attach_when_http_error(self, respx_mock):
        # given the server returns an error
        respx_mock.get("/access/api/roles/cd1?scope=sc1").mock(return_value=httpx.Response(500, json={}))
        client = self.client
        sut = access.RoleRef(id="one", scope="sc1", code="cd1")
        # when we call attach an exception is raised
        with pytest.raises(httpx.HTTPError):
            sut.attach(client)


@pytest.mark.respx(base_url=TEST_BASE)
class DescribeRole:
    client = httpx.Client(base_url=TEST_BASE, event_hooks={"response": [response_hook]})

    @pytest.fixture
    def stock_policy_resource(self):
        return access.PolicyResource(
            id="policy1",
            code="a_policy",
            description="policy description",
            applications=["lusid"],
            grant=access.Grant.ALLOW,
            when=access.WhenSpec(activate="2024-08-31T18:00:00.0000000+00:00"),
            selectors=[
                access.IdSelector(
                    name="banjo",
                    description="whatever",
                    identifier={"scope": "w", "code": "y"},
                    actions=[access.ActionId(scope="actscope", activity="execute", entity="Feature")],
                )
            ],
        )

    def test_create_with_new_resource_request(self, respx_mock, stock_policy_resource):
        respx_mock.post("/access/api/roles").mock(return_value=httpx.Response(200, json={}))
        # given no role exists yet and we create one using the updadted RoleResourceRequest
        sut = access.RoleResource(
            id="role1",
            code="a_role_code",
            resource=access.RoleResourceRequest(
                policy_id_role_resource=access.PolicyIdRoleResource(policies=[stock_policy_resource])
            ),
            when=access.WhenSpec(activate="2024-03-01:00:00.0000000+00:00"),
            permission=access.Permission.READ,
        )
        sut.create(self.client)
        # then a post request is made
        request = respx_mock.calls.last.request
        assert request.method == "POST"
        assert request.url.path == "/access/api/roles"
        assert json.loads(request.content) == {
            "code": "a_role_code",
            "permission": "Read",
            "resource": {
                "policyIdRoleResource": {
                    "policies": [{"code": "a_policy", "scope": "default"}],
                    "policyCollections": [],
                }
            },
            "when": {"activate": "2024-03-01:00:00.0000000+00:00"},
        }

    def test_can_parse_role_with_policy_id(self):
        serialized = {
            "id": "whatever",
            "code": "a_role_code",
            "permission": "Read",
            "resource": {
                "policyIdRoleResource": {
                    "policies": [],
                    "policyCollections": [],
                }
            },
            "when": {"activate": "2024-03-01:00:00.0000000+00:00"},
        }
        sut = access.RoleResource.model_validate(serialized, context={"$refs": {}})
        assert sut.resource
        assert sut.resource.policy_id_role_resource is not None

    def test_can_parse_role_with_non_transitive(self):
        serialized = {
            "id": "whatever",
            "code": "a_role_code",
            "permission": "Read",
            "resource": {
                "nonTransitiveSupervisorRoleResource": {
                    "roles": [],
                }
            },
            "when": {"activate": "2024-03-01:00:00.0000000+00:00"},
        }
        sut = access.RoleResource.model_validate(serialized, context={"$refs": {}})
        assert sut.resource
        assert sut.resource.non_transitive_supervisor_role_resource is not None

    def test_create_with_nontransitive(self, respx_mock, stock_policy_resource):
        respx_mock.post("/access/api/roles").mock(return_value=httpx.Response(200, json={}))
        # given a pre-existing supervisor role
        sup = access.RoleResource(
            id="sup",
            code="supervisor",
            resource=access.RoleResourceRequest(
                policy_id_role_resource=access.PolicyIdRoleResource(policies=[])
            ),
            when=access.WhenSpec(activate="2024-03-01:00:00.0000000+00:00"),
            permission=access.Permission.READ,
        )
        # when we create a new role which references the supervisor
        sut = access.RoleResource(
            id="role1",
            code="a_role_code",
            resource=access.RoleResourceRequest(
                non_transitive_supervisor_role_resource=access.NonTransitiveSupervisorRoleResource(
                    roles=[sup]
                )
            ),
            when=access.WhenSpec(activate="2024-03-01:00:00.0000000+00:00"),
            permission=access.Permission.READ,
        )
        sut.create(self.client)
        # then a post request is made
        request = respx_mock.calls.last.request
        assert request.method == "POST"
        assert request.url.path == "/access/api/roles"
        assert json.loads(request.content) == {
            "code": "a_role_code",
            "permission": "Read",
            "resource": {
                "nonTransitiveSupervisorRoleResource": {
                    "roles": [{"scope": "default", "code": "supervisor"}]
                }
            },
            "when": {"activate": "2024-03-01:00:00.0000000+00:00"},
        }

    def test_create_with_policy_collection(self, respx_mock, stock_policy_resource):
        respx_mock.post("/access/api/roles").mock(return_value=httpx.Response(200, json={}))
        respx_mock.post("/access/api/policycollections").mock(
            return_value=httpx.Response(200, json={"id": {"scope": "default", "code": "col-inner"}})
        )
        respx_mock.post("/access/api/policycollections").mock(
            return_value=httpx.Response(200, json={"id": {"scope": "default", "code": "col-code"}})
        )
        # given an inner collection which has already been created
        inner = access.PolicyCollectionResource(id="polly-collection", code="col-code", policies=[])
        inner.create(self.client)
        # when we create a role referenccing the collection
        sut = access.RoleResource(
            id="role1",
            code="a_role_code",
            resource=access.RoleResourceRequest(
                policy_id_role_resource=access.PolicyIdRoleResource(policy_collections=[inner])
            ),
            when=access.WhenSpec(activate="2024-03-01:00:00.0000000+00:00"),
            permission=access.Permission.READ,
        )
        sut.create(self.client)
        # then a post request is made
        request = respx_mock.calls.last.request
        assert request.method == "POST"
        assert request.url.path == "/access/api/roles"
        assert json.loads(request.content) == {
            "code": "a_role_code",
            "permission": "Read",
            "resource": {
                "policyIdRoleResource": {
                    "policies": [],
                    "policyCollections": [{"scope": "default", "code": "col-code"}],
                }
            },
            "when": {"activate": "2024-03-01:00:00.0000000+00:00"},
        }

    def test_delete(self, respx_mock, stock_policy_resource):
        respx_mock.delete("/access/api/roles/a_role").mock(side_effect=[httpx.Response(200, json={})])
        # given an existing policy
        # when we delete it
        old_state = SimpleNamespace(id="role1", code="a_role")
        access.RoleResource.delete(self.client, old_state)
        # then a delete request is made
        request = respx_mock.calls.last.request
        assert request.method == "DELETE"
        assert request.url.path == "/access/api/roles/a_role"

    @staticmethod
    def test_deps(stock_policy_resource):
        # given a role with two policy resources
        sut = access.RoleResource(
            id="role1",
            code="a_role_code",
            resource=access.RoleResourceRequest(
                policy_id_role_resource=access.PolicyIdRoleResource(policies=[stock_policy_resource])
            ),
            when=access.WhenSpec(activate="2024-03-01:00:00.0000000+00:00"),
            permission=access.Permission.READ,
        )
        # when we ask for the deps we get two back
        assert sut.deps() == [stock_policy_resource]

    def test_deps_policy_collection(self, stock_policy_resource):
        # given a role with a policy collection
        inner = access.PolicyCollectionResource(
            id="polly-collection", code="col-inner", policies=[stock_policy_resource]
        )
        sut = access.RoleResource(
            id="role1",
            code="a_role_code",
            resource=access.RoleResourceRequest(
                policy_id_role_resource=access.PolicyIdRoleResource(policy_collections=[inner])
            ),
            when=access.WhenSpec(activate="2024-03-01:00:00.0000000+00:00"),
            permission=access.Permission.READ,
        )
        # when we ask for the deps
        assert sut.deps() == [inner]

    @pytest.fixture
    def read_role_response(self):
        return {
            "id": {"scope": "default", "code": "a_role_code"},
            "when": {
                "activate": "2024-03-01:00:00.0000000+00:00",
                "deactivate": "2024-03-01:00:00.0000000+00:00",
            },
            "roleHierarchyIndex": 20,
            "permission": "Read",
            "resource": {
                "policyIdRoleResource": {"policies": [{"code": "a_policy", "scope": "default"}]}
            },
            "links": [],
        }

    def test_update_remove_policies(self, respx_mock, read_role_response):
        # given the role exists and has a policy
        respx_mock.get("/access/api/roles/a_role_code").mock(
            side_effect=[httpx.Response(200, json=read_role_response)]
        )
        respx_mock.put("/access/api/roles/a_role_code").mock(side_effect=[httpx.Response(200, json={})])
        # and the desired state is no policies
        sut = access.RoleResource(
            id="role1",
            code="a_role_code",
            resource=access.RoleResourceRequest(
                policy_id_role_resource=access.PolicyIdRoleResource(policies=[])
            ),
            when=access.WhenSpec(activate="2024-03-01:00:00.0000000+00:00"),
            permission=access.Permission.READ,
        )
        # when we update
        log = sut.update(self.client, SimpleNamespace(id="role1", code="a_role_code"))
        # then the log returned is the same
        assert log == {"code": "a_role_code", "id": "role1"}
        # and a put request to modify the role was made
        request = respx_mock.calls.last.request
        assert request.method == "PUT"
        assert request.url.path == "/access/api/roles/a_role_code"
        assert json.loads(request.content) == {
            "permission": "Read",  # nb no code set when its a PUT
            "resource": {"policyIdRoleResource": {"policies": [], "policyCollections": []}},
            "when": {"activate": "2024-03-01:00:00.0000000+00:00"},
        }

    def test_update_non_transitive(self, respx_mock, read_role_response):
        # given the role exists and has a policy
        respx_mock.get("/access/api/roles/a_role_code").mock(
            side_effect=[httpx.Response(200, json=read_role_response)]
        )
        respx_mock.put("/access/api/roles/a_role_code").mock(side_effect=[httpx.Response(200, json={})])
        # and there is an existing supervisor role
        sup = access.RoleResource(
            id="sup",
            code="supervisor",
            resource=access.RoleResourceRequest(
                policy_id_role_resource=access.PolicyIdRoleResource(policies=[])
            ),
            when=access.WhenSpec(activate="2024-03-01:00:00.0000000+00:00"),
            permission=access.Permission.READ,
        )
        # and the desired state only a nonTransitiveSupervisor....
        sut = access.RoleResource(
            id="role1",
            code="a_role_code",
            resource=access.RoleResourceRequest(
                non_transitive_supervisor_role_resource=access.NonTransitiveSupervisorRoleResource(
                    roles=[sup]
                )
            ),
            when=access.WhenSpec(activate="2024-03-01:00:00.0000000+00:00"),
            permission=access.Permission.READ,
        )
        # when we update
        log = sut.update(self.client, SimpleNamespace(id="role1", code="a_role_code"))
        # then the log returned is the same
        assert log == {"code": "a_role_code", "id": "role1"}
        # and a put request to modify the role was made
        request = respx_mock.calls.last.request
        assert request.method == "PUT"
        assert request.url.path == "/access/api/roles/a_role_code"
        assert json.loads(request.content) == {
            "permission": "Read",  # nb no code set when its a PUT
            "resource": {
                "nonTransitiveSupervisorRoleResource": {
                    "roles": [{"scope": "default", "code": "supervisor"}]
                }
            },
            "when": {"activate": "2024-03-01:00:00.0000000+00:00"},
        }

    def test_update_non_transitive_ref(self, respx_mock, read_role_response):
        # given the role exists and has a policy
        respx_mock.get("/access/api/roles/a_role_code").mock(
            side_effect=[httpx.Response(200, json=read_role_response)]
        )
        respx_mock.put("/access/api/roles/a_role_code").mock(side_effect=[httpx.Response(200, json={})])
        # and we reference a system role
        ref = access.RoleRef(id="supervisor", scope="support", code="support-iam-readonly")
        # and the desired state only a nonTransitiveSupervisor....
        ntsrs = access.NonTransitiveSupervisorRoleResource(roles=[ref])
        sut = access.RoleResource(
            id="role1",
            code="a_role_code",
            resource=access.RoleResourceRequest(non_transitive_supervisor_role_resource=ntsrs),
            when=access.WhenSpec(activate="2024-03-01:00:00.0000000+00:00"),
            permission=access.Permission.READ,
        )
        # when we update
        log = sut.update(self.client, SimpleNamespace(id="role1", code="a_role_code"))
        # then the log returned is the same
        assert log == {"code": "a_role_code", "id": "role1"}
        # and a put request to modify the role was made
        request = respx_mock.calls.last.request
        assert request.method == "PUT"
        assert request.url.path == "/access/api/roles/a_role_code"
        assert json.loads(request.content) == {
            "permission": "Read",  # nb no code set when its a PUT
            "resource": {
                "nonTransitiveSupervisorRoleResource": {
                    "roles": [{"scope": "support", "code": "support-iam-readonly"}]
                }
            },
            "when": {"activate": "2024-03-01:00:00.0000000+00:00"},
        }

    def test_dump(self, stock_policy_resource):
        # given a role resource with policies
        sut = access.RoleResource(
            id="role1",
            code="test-role",
            description="A test role with policies",
            resource=access.RoleResourceRequest(
                policy_id_role_resource=access.PolicyIdRoleResource(
                    policies=[stock_policy_resource]
                )
            ),
            when=access.WhenSpec(activate="2024-01-01T00:00:00.0000000+00:00"),
            permission=access.Permission.READ
        )
        # when we dump it
        dumped = sut.model_dump(
            mode="json",
            by_alias=True,
            round_trip=True,
            exclude_none=True,
            context={"style": "dump"}
        )
        # then the dumped state is correct
        assert dumped == {
            "code": "test-role",
            "description": "A test role with policies",
            "resource": {
                "policyIdRoleResource": {
                    "policies": [
                        {"$ref": "policy1"}
                    ],
                    "policyCollections": []
                }
            },
            "when": {
                "activate": "2024-01-01T00:00:00.0000000+00:00"
            },
            "permission": "Read"
        }

    def test_undump(self, stock_policy_resource):
        # given a dumped role state
        dumped = {
            "code": "test-role",
            "description": "A test role with policies",
            "resource": {
                "policyIdRoleResource": {
                    "policies": [
                        {"$ref": "policy1"}
                    ],
                    "policyCollections": []
                }
            },
            "when": {
                "activate": "2024-01-01T00:00:00.0000000+00:00"
            },
            "permission": "Read"
        }
        # when we undump it
        sut = access.RoleResource.model_validate(
            dumped,
            context={
                "style": "undump",
                "$refs": {
                    "policy1": stock_policy_resource,
                },
                "id": "role1",
            }
        )
        # then the id has been extracted from the context
        assert sut.id == "role1"
        assert sut.code == "test-role"
        assert sut.description == "A test role with policies"
        assert sut.permission == access.Permission.READ
        assert sut.when.activate == "2024-01-01T00:00:00.0000000+00:00"
        # and the policy references have been wired up
        assert sut.resource
        assert sut.resource.policy_id_role_resource is not None
        assert sut.resource.policy_id_role_resource.policies is not None
        assert len(sut.resource.policy_id_role_resource.policies) == 1
        assert sut.resource.policy_id_role_resource.policies[0] == stock_policy_resource
        assert sut.resource.policy_id_role_resource.policy_collections is not None
        assert len(sut.resource.policy_id_role_resource.policy_collections) == 0


@pytest.mark.respx(base_url=TEST_BASE)
class DescribePolicyCollectionRef:
    client = httpx.Client(base_url=TEST_BASE, event_hooks={"response": [response_hook]})

    def test_attach_when_present(self, respx_mock):
        # given that the remote definition exists
        respx_mock.get("/access/api/policycollections/cd1?scope=sc1").mock(
            return_value=httpx.Response(200, json={})
        )
        client = self.client
        sut = access.PolicyCollectionRef(id="one", scope="sc1", code="cd1")
        # when we call attach
        sut.attach(client)
        # then a get request was made and no exception raised

    def test_attach_when_missing(self, respx_mock):
        # given the remote does not exist
        respx_mock.get("/access/api/policycollections/cd1?scope=sc1").mock(
            return_value=httpx.Response(404, json={})
        )
        client = self.client
        sut = access.PolicyCollectionRef(id="one", scope="sc1", code="cd1")
        # when we call attach an exception is raised
        with pytest.raises(RuntimeError) as ex:
            sut.attach(client)
        assert "PolicyCollection sc1/cd1 not found" in str(ex.value)

    def test_attach_when_http_error(self, respx_mock):
        # given the server returns an error
        respx_mock.get("/access/api/policycollections/cd1?scope=sc1").mock(
            return_value=httpx.Response(500, json={})
        )
        client = self.client
        sut = access.PolicyCollectionRef(id="one", scope="sc1", code="cd1")
        # when we call attach an exception is raised
        with pytest.raises(httpx.HTTPError):
            sut.attach(client)


@pytest.mark.respx(base_url=TEST_BASE)
class DescribePolicyCollectionResource:
    client = httpx.Client(base_url=TEST_BASE, event_hooks={"response": [response_hook]})

    @pytest.fixture
    def policy1(self):
        return access.PolicyResource(
            id="policy1",
            code="a_policy1",
            description="policy description",
            applications=["lusid"],
            grant=access.Grant.ALLOW,
            when=access.WhenSpec(activate="2024-08-31T18:00:00.0000000+00:00"),
            selectors=[
                access.IdSelector(
                    name="banjo",
                    description="whatever",
                    identifier={"scope": "w", "code": "y"},
                    actions=[access.ActionId(scope="actscope", activity="execute", entity="Feature")],
                )
            ],
        )

    @pytest.fixture
    def policy2(self):
        return access.PolicyResource(
            id="policy2",
            code="a_policy2",
            description="policy description",
            applications=["lusid"],
            grant=access.Grant.ALLOW,
            when=access.WhenSpec(activate="2024-08-31T18:00:00.0000000+00:00"),
            selectors=[
                access.MatchAllSelector(
                    name="banjo",
                    description="whatever",
                    actions=[access.ActionId(scope="actscope", activity="execute", entity="Feature")],
                )
            ],
        )

    def test_create_with_policies(self, respx_mock, policy1, policy2):
        respx_mock.post("/access/api/policycollections").mock(
            return_value=httpx.Response(200, json={"id": {"scope": "default", "code": "col-code"}})
        )
        client = self.client
        # given a desired collection
        sut = access.PolicyCollectionResource(
            id="polly-collection", code="col-code", policies=[policy1, policy2]
        )
        # when we create it
        state = sut.create(client)
        # then the state is returned
        assert state == {"scope": "default", "code": "col-code"}
        # and a create request was sent
        request = respx_mock.calls.last.request
        assert request.method == "POST"
        assert request.url.path == "/access/api/policycollections"
        assert json.loads(request.content) == {
            "code": "col-code",
            "policies": [
                {"scope": "default", "code": "a_policy1"},
                {"scope": "default", "code": "a_policy2"},
            ],
            "policyCollections": [],
        }

    def test_create_with_policy_ref(self, respx_mock, policy1, policy2):
        respx_mock.post("/access/api/policycollections").mock(
            return_value=httpx.Response(200, json={"id": {"scope": "default", "code": "col-code"}})
        )
        client = self.client
        # given a policy ref to a non-default scope (eg a ref to a builtin
        # policy like an admin one
        ref = access.PolicyRef(id="ref", scope="sc1", code="cd1")
        # given a desired collection
        sut = access.PolicyCollectionResource(id="polly-collection", code="col-code", policies=[ref])
        # when we create it
        state = sut.create(client)
        # then the state is returned
        assert state == {"scope": "default", "code": "col-code"}
        # and a create request was sent
        request = respx_mock.calls.last.request
        assert request.method == "POST"
        assert request.url.path == "/access/api/policycollections"
        assert json.loads(request.content) == {
            "code": "col-code",
            "policies": [{"scope": "sc1", "code": "cd1"}],
            "policyCollections": [],
        }

    def test_create_with_nested_collection(self, respx_mock, policy1, policy2):
        respx_mock.post("/access/api/policycollections").mock(
            return_value=httpx.Response(200, json={"id": {"scope": "default", "code": "col-inner"}})
        )
        respx_mock.post("/access/api/policycollections").mock(
            return_value=httpx.Response(200, json={"id": {"scope": "default", "code": "col-code"}})
        )
        client = self.client
        # given an inner collection which has already been created
        inner = access.PolicyCollectionResource(
            id="polly-collection", code="col-inner", policies=[policy1, policy2]
        )
        inner.create(self.client)
        # and a desired collection which references it
        sut = access.PolicyCollectionResource(
            id="polly-collection", code="col-code", policy_collections=[inner]
        )
        # when we create it
        state = sut.create(client)
        # then the state is returned
        assert state == {"scope": "default", "code": "col-code"}
        # and a create request was sent
        request = respx_mock.calls.last.request
        assert request.method == "POST"
        assert request.url.path == "/access/api/policycollections"
        assert json.loads(request.content) == {
            "code": "col-code",
            "policyCollections": [{"scope": "default", "code": "col-inner"}],
            "policies": [],
        }

    def test_update_no_change(self, respx_mock, policy1, policy2):
        # given an existing collection
        respx_mock.get("/access/api/policycollections/col-code").mock(
            return_value=httpx.Response(
                200,
                json={
                    "id": {"code": "col-code", "scope": "default"},
                    "policies": [
                        {"scope": "default", "code": "a_policy1"},
                        {"scope": "default", "code": "a_policy2"},
                    ],
                    "policyCollections": [],
                    "links": [],
                },
            )
        )
        old_state = SimpleNamespace(scope="default", code="col-code")
        # given a desired collection
        sut = access.PolicyCollectionResource(
            id="polly-collection", code="col-code", policies=[policy1, policy2]
        )
        # when we update
        state = sut.update(self.client, old_state)
        # then state is None and no further requests are made
        assert state is None

    def test_update_rename(self, respx_mock, policy1, policy2):
        respx_mock.delete("/access/api/policycollections/col-code").mock(
            side_effect=[httpx.Response(200, json={})]
        )
        respx_mock.post("/access/api/policycollections").mock(
            return_value=httpx.Response(200, json={"id": {"scope": "default", "code": "col-code2"}})
        )
        # given an existing collection at col-code
        old_state = SimpleNamespace(scope="default", code="col-code")
        # and a desired collection with code col-code2
        sut = access.PolicyCollectionResource(
            id="polly-collection", code="col-code2", policies=[policy1, policy2]
        )
        # when we update
        state = sut.update(self.client, old_state)
        # the new state is returned
        assert state == {"scope": "default", "code": "col-code2"}
        # and a delete and a create call were made

    def test_update_remove_policy(self, respx_mock, policy1, policy2):
        # given an existing collection
        respx_mock.get("/access/api/policycollections/col-code").mock(
            return_value=httpx.Response(
                200,
                json={
                    "id": {"code": "col-code", "scope": "default"},
                    "policies": [
                        {"scope": "default", "code": "a_policy1"},
                        {"scope": "default", "code": "a_policy2"},
                    ],
                    "policyCollections": [],
                    "links": [],
                },
            )
        )
        respx_mock.put("/access/api/policycollections/col-code").mock(
            return_value=httpx.Response(200, json={"id": {"scope": "default", "code": "col-code"}})
        )
        old_state = SimpleNamespace(scope="default", code="col-code")
        # given a desired collection
        sut = access.PolicyCollectionResource(id="polly-collection", code="col-code", policies=[policy1])
        # when we update
        state = sut.update(self.client, old_state)
        # then state returned
        assert state == {"scope": "default", "code": "col-code"}
        # and a put request was sent
        request = respx_mock.calls.last.request
        assert request.method == "PUT"
        assert request.url.path == "/access/api/policycollections/col-code"
        assert json.loads(request.content) == {
            "policies": [{"scope": "default", "code": "a_policy1"}],
            "policyCollections": [],
        }

    def test_delete(self, respx_mock):
        respx_mock.delete("/access/api/policycollections/col-code").mock(
            side_effect=[httpx.Response(200, json={})]
        )
        # given an existing collection
        # when we delete it
        old_state = SimpleNamespace(scope="default", code="col-code")
        access.PolicyCollectionResource.delete(self.client, old_state)
        # then a delete request is made
        request = respx_mock.calls.last.request
        assert request.method == "DELETE"
        assert request.url.path == "/access/api/policycollections/col-code"

    def test_deps_policy_ref(self, respx_mock, policy1, policy2):
        ref = access.PolicyRef(id="ref", scope="sc1", code="cd1")
        # given a desired collection with two policies
        sut = access.PolicyCollectionResource(id="polly-collection", code="col-code", policies=[ref])
        # when we get the deps
        deps = sut.deps()
        assert deps == [ref]

    def test_deps_policies_only(self, respx_mock, policy1, policy2):
        # given a desired collection with two policies
        sut = access.PolicyCollectionResource(
            id="polly-collection", code="col-code", policies=[policy1, policy2]
        )
        # when we get the deps
        deps = sut.deps()
        assert deps == [policy1, policy2]

    def test_deps_collections_only(self, respx_mock, policy1, policy2):
        # given an inner collection which has already been created
        inner = access.PolicyCollectionResource(
            id="polly-collection", code="col-inner", policies=[policy1, policy2]
        )
        # and a desired collection which references it
        sut = access.PolicyCollectionResource(
            id="polly-collection", code="col-code", policy_collections=[inner]
        )
        # when we get the deps
        deps = sut.deps()
        assert deps == [inner]

    def test_dump(self, policy1, policy2):
        # given an existing policy collection
        sut = access.PolicyCollectionResource(
            id="collection1",
            code="test-collection",
            description="A test collection",
            policies=[policy1, policy2],
            metadata={"key1": ["value1", "value2"], "key2": ["value3"]}
        )
        # when we dump it
        dumped = sut.model_dump(
            by_alias=True,
            round_trip=True,   # excludes computeds
            exclude_none=True,
            context={"style": "dump"}
        )
        # then the dumped state is correct
        assert dumped == {
            "code": "test-collection",
            "description": "A test collection",
            "policies": [
                {"$ref": policy1.id},
                {"$ref": policy2.id}
            ],
            "policyCollections": [],
            "metadata": {
                "key1": ["value1", "value2"],
                "key2": ["value3"]
            }
        }

    def test_undump(self, policy1, policy2):
        # given a dumped policy collection state
        dumped = {
            "code": "test-collection",
            "description": "A test collection",
            "policies": [
                {"$ref": policy1.id},
                {"$ref": policy2.id}
            ],
            "policyCollections": [],
            "metadata": {
                "key1": ["value1", "value2"],
                "key2": ["value3"]
            }
        }
        # when we undump it
        sut = access.PolicyCollectionResource.model_validate(
            dumped,
            context={
                "style": "undump",
                "$refs": {
                    policy1.id: policy1,
                    policy2.id: policy2,
                },
                "id": "collection1",
            }
        )
        # then the id has been extracted from the context
        assert sut.id == "collection1"
        assert sut.code == "test-collection"
        assert sut.description == "A test collection"
        # and the policy refs have been wired up
        assert sut.policies == [policy1, policy2]
        assert sut.policy_collections == []
        assert sut.metadata == {
            "key1": ["value1", "value2"],
            "key2": ["value3"]
        }

    def test_parse_api_format(self):
        policy1 = access.PolicyRef(
            id="p1",
            scope="sc1",
            code="cd1",
        )
        coll1 = access.PolicyCollectionRef(
            id="c1",
            scope="sc1",
            code="cd1"
        )
        api_format = {
          "id": {
            "scope": "default",
            "code": "organisation-wide-policies"
          },
          "policies": [
              {"$ref": "p1"}
          ],
          "policyCollections": [
              {"$ref": "c1"}
          ],
          "description": "Collection of organisation wide policies"
        }
        parsed = access.PolicyCollectionResource.model_validate(
            api_format,
            context={"id": "someid", "$refs": {
                "p1": policy1,
                "c1": coll1
            }}
        )
        assert parsed.id == "someid"
        assert parsed.policies == [policy1]
        assert parsed.policy_collections == [coll1]


@pytest.mark.respx(base_url=TEST_BASE)
class DescribePolicyTemplateRef:
    client = httpx.Client(base_url=TEST_BASE, event_hooks={"response": [response_hook]})

    @staticmethod
    def test_scope_default():
        # when scope is omitted
        sut = access.PolicyTemplateRef(id="one", code="cd")
        # then it uses "default"
        assert sut.scope == "default"

    def test_attach_when_present(self, respx_mock):
        # given that the remote definition exists
        respx_mock.get("/access/api/policytemplates/pol1?scope=sc1").mock(
            return_value=httpx.Response(200, json={})
        )
        client = self.client
        sut = access.PolicyTemplateRef(id="one", scope="sc1", code="pol1")

        # when we call attach
        sut.attach(client)
        # then a get request was made and no exception raised

    def test_attach_when_missing(self, respx_mock):
        # given the remote does not exist
        respx_mock.get("/access/api/policytemplates/pol1?scope=sc1").mock(
            return_value=httpx.Response(404, json={})
        )
        client = self.client
        sut = access.PolicyTemplateRef(id="one", scope="sc1", code="pol1")
        # when we call attach an exception is raised
        with pytest.raises(RuntimeError) as ex:
            sut.attach(client)
        assert "Policy Template sc1/pol1 not found" in str(ex.value)

    def test_attach_when_http_error(self, respx_mock):
        # given the server returns an error
        respx_mock.get("/access/api/policytemplates/pol1?scope=sc1").mock(
            return_value=httpx.Response(500, json={})
        )
        client = self.client
        sut = access.PolicyTemplateRef(id="one", scope="sc1", code="pol1")
        # when we call attach an exception is raised
        with pytest.raises(httpx.HTTPError):
            sut.attach(client)


@pytest.mark.respx(base_url=TEST_BASE)
class DescribePolicyTemplate:
    client = httpx.Client(base_url=TEST_BASE, event_hooks={"response": [response_hook]})

    @pytest.fixture
    def policy1(self):
        return access.PolicyResource(
            id="policy1",
            code="a_policy1",
            description="policy description",
            applications=["lusid"],
            grant=access.Grant.ALLOW,
            when=access.WhenSpec(activate="2024-08-31T18:00:00.0000000+00:00"),
            selectors=[
                access.IdSelector(
                    name="banjo",
                    description="whatever",
                    identifier={"scope": "w", "code": "y"},
                    actions=[access.ActionId(scope="actscope", activity="execute", entity="Feature")],
                )
            ],
        )

    @pytest.fixture
    def policy2(self):
        return access.PolicyResource(
            id="policy2",
            code="a_policy2",
            description="policy description",
            applications=["lusid"],
            grant=access.Grant.ALLOW,
            when=access.WhenSpec(activate="2024-08-31T18:00:00.0000000+00:00"),
            selectors=[
                access.MatchAllSelector(
                    name="banjo",
                    description="whatever",
                    actions=[access.ActionId(scope="actscope", activity="execute", entity="Feature")],
                )
            ],
        )

    @pytest.fixture
    def stock_policy_template_resource(self):
        selector = access.IdSelector(
            name="example_id_selector",
            description="test_description",
            identifier={"scope": "w"},
            actions=[access.ActionId(scope="default", activity="Read", entity="Portfolio")],
        )
        return access.PolicyTemplateResource(
            id="policy_template1",
            display_name="display_name",
            code="code1",
            description="some_description",
            templated_selectors=[access.TemplatedSelector(
                application="LUSID",
                tag="Data",
                selector=selector
            )]
        )

    def test_read(self, respx_mock, stock_policy_template_resource):
        respx_mock.get("/access/api/policytemplates/code1").mock(
            return_value=httpx.Response(
                200,
                json={
                    "displayName": "display_name",
                    "scope": "scope1",
                    "code": "code1",
                    "description": "some_description",
                    "applications": [
                        "LUSID"
                    ],
                    "tags": [
                        "example_tag"
                    ],
                    "templatedSelectors": [
                        {
                        "application": "LUSID",
                        "tag": "example_tag",
                        "selector": {
                            "idSelectorDefinition": {
                            "identifier": {
                                "scope": "w"
                            },
                            "actions": [
                                {
                                "scope": "default",
                                "activity": "Read",
                                "entity": "Portfolio"
                                },
                            ],
                            "name": "example_id_selector",
                            "description": "test_description"
                            }
                        }
                        }
                    ]
                }
            )
        )
        sut = stock_policy_template_resource
        old_state = SimpleNamespace(code=stock_policy_template_resource.code)
        result = sut.read(self.client, old_state)
        assert result is not None
        assert result["code"] == "code1"
        assert result["scope"] == "scope1"
        assert result["displayName"] == "display_name"

    def test_create(self, respx_mock, stock_policy_template_resource):
        respx_mock.post("/access/api/policytemplates").mock(return_value=httpx.Response(200, json={}))
        sut = stock_policy_template_resource
        result = sut.create(self.client)
        assert result is not None
        assert result["id"] == "policy_template1"
        assert result["code"] == "code1"
        request = respx_mock.calls.last.request
        assert request.method == "POST"
        assert request.url.path == "/access/api/policytemplates"
        assert json.loads(request.content) == {
            "code": "code1",
            "displayName": "display_name",
            "description": "some_description",
            "templatedSelectors": [
                {
                "application": "LUSID",
                "tag": "Data",
                "selector": {
                    "idSelectorDefinition": {
                    "identifier": {
                        "scope": "w"
                    },
                    "actions": [
                        {
                        "scope": "default",
                        "activity": "Read",
                        "entity": "Portfolio"
                        },
                    ],
                    "name": "example_id_selector",
                    "description": "test_description"
                    }
                }
                }
            ]
        }
        assert "source_version" in result
        assert "remote_version" in result

    def test_update_no_change(self, respx_mock, stock_policy_template_resource):
        sut = stock_policy_template_resource
        remote_response = {
            "displayName": "display_name",
            "scope": "scope1",
            "code": "code1",
            "description": "some_description",
            "applications": [
                "LUSID"
            ],
            "tags": [
                "example_tag"
            ],
            "templatedSelectors": [
                {
                "application": "LUSID",
                "tag": "example_tag",
                "selector": {
                    "idSelectorDefinition": {
                        "identifier": {
                            "scope": "w"
                        },
                        "actions": [
                            {
                            "scope": "default",
                            "activity": "Read",
                            "entity": "Portfolio"
                            },
                        ],
                        "name": "example_id_selector",
                        "description": "test_description"
                        }
                    }
                }
            ]
        }
        respx_mock.get("/access/api/policytemplates/code1").mock(
            return_value=httpx.Response(200, json=remote_response)
        )
        # Calculate hashes to simulate no change scenario
        source_hash = sut.__get_content_hash__()
        # The remote hash should match what update() calculates from the read() response
        remote_hash = sha256(json.dumps(remote_response, sort_keys=True).encode()).hexdigest()
        old_state = SimpleNamespace(
            code="code1", source_version=source_hash, remote_version=remote_hash
        )
        result = sut.update(self.client, old_state)
        assert result is None

    def test_update_with_change(self, respx_mock):
        respx_mock.get("/access/api/policytemplates/code1").mock(
            return_value=httpx.Response(
                200,
                json={
                    "displayName": "display_name",
                    "scope": "scope1",
                    "code": "code1",
                    "description": "some_description",
                    "applications": [
                        "LUSID"
                    ],
                    "tags": [
                        "example_tag"
                    ],
                    "templatedSelectors": [
                        {
                        "application": "LUSID",
                        "tag": "example_tag",
                        "selector": {
                            "idSelectorDefinition": {
                                "identifier": {
                                    "scope": "w"
                                },
                                "actions": [
                                    {
                                    "scope": "default",
                                    "activity": "Read",
                                    "entity": "Portfolio"
                                    },
                                ],
                                "name": "example_id_selector",
                                "description": "test_description"
                                }
                            }
                        }
                    ]
                }
            )
        )
        respx_mock.put("/access/api/policytemplates/code1").mock(
            return_value=httpx.Response(
                200,
                json={
                    "displayName": "display_name_another",
                    "scope": "scope2",
                    "code": "code1",
                    "description": "some_description_another",
                    "applications": [
                        "LUSID"
                    ],
                    "tags": [
                        "example_tag"
                    ],
                    "templatedSelectors": [
                        {
                        "application": "LUSID",
                        "tag": "example_tag",
                        "selector": {
                            "idSelectorDefinition": {
                            "identifier": {
                                "scope": "w"
                            },
                            "actions": [
                                {
                                "scope": "default",
                                "activity": "Read",
                                "entity": "Portfolio"
                                },
                            ],
                            "name": "example_id_selector",
                            "description": "test_description"
                            }
                        }
                        }
                    ]
                }
            )
        )
        id_selector = access.IdSelector(
            name="example_id_selector",
            description="test_description",
            identifier={"scope": "w"},
            actions=[access.ActionId(scope="default", activity="Read", entity="Portfolio")],
        )
        templated_selector = access.TemplatedSelector(
                application="LUSID",
                tag="example_tag",
                selector=id_selector
            )
        sut = access.PolicyTemplateResource(
            id="policy_template1",
            display_name="display_name_another",
            code="code1",
            description="some_description_another",
            templated_selectors=[templated_selector]
        )
        old_state = SimpleNamespace(
            code="code1", remote_version="oldhash", source_version="different_hash"
        )
        result = sut.update(self.client, old_state)
        assert result is not None
        assert result["id"] == "policy_template1"
        assert result["code"] == "code1"
        assert "source_version" in result
        assert "remote_version" in result
        # Check for put request is called
        request = respx_mock.calls.last.request
        assert request.method == "PUT"
        assert request.url.path == "/access/api/policytemplates/code1"
        assert json.loads(request.content) == {
            "displayName": "display_name_another",
            "description": "some_description_another",
            "templatedSelectors": [
                {
                "application": "LUSID",
                "tag": "example_tag",
                "selector": {
                    "idSelectorDefinition": {
                    "identifier": {
                        "scope": "w"
                    },
                    "actions": [
                        {
                        "scope": "default",
                        "activity": "Read",
                        "entity": "Portfolio"
                        },
                    ],
                    "name": "example_id_selector",
                    "description": "test_description"
                    }
                }
                }
            ]
            }

    def test_delete(self, respx_mock):
        respx_mock.delete("/access/api/policytemplates/code1").mock(
            return_value=httpx.Response(200, json={})
        )
        old_state = SimpleNamespace(code="code1")
        access.PolicyTemplateResource.delete(self.client, old_state)
        request = respx_mock.calls[-1].request
        assert request.method == "DELETE"
        assert request.url.path == "/access/api/policytemplates/code1"

    def test_deps(self, stock_policy_template_resource):
        sut = stock_policy_template_resource
        assert sut.deps() == []

    def test_dump(self, stock_policy_template_resource):
        sut = stock_policy_template_resource
        dumped = sut.model_dump(
            by_alias=True,
            round_trip=True,   # excludes computeds
            exclude_none=True,
            context={"style": "dump"}
        )
        # then the dumped state is correct
        assert dumped == {
           "code": "code1",
           "description": "some_description",
           "displayName": "display_name",
           "templatedSelectors": [{
               "application": "LUSID",
               "selector": {
                   "actions": [{
                       "activity": "Read",
                       "entity": "Portfolio",
                       "scope": "default",
                   }],
                   "description": "test_description",
                   "identifier": {
                       "scope": "w",
                   },
                   "name": "example_id_selector",
                   "typeName": "idSelectorDefinition",
               },
               "tag": "Data",
           }],
         }

    def test_undump(self):
        # given a dump
        data = {
           "code": "code1",
           "description": "some_description",
           "displayName": "display_name",
           "templatedSelectors": [{
               "application": "LUSID",
               "selector": {
                   "actions": [{
                       "activity": "Read",
                       "entity": "Portfolio",
                       "scope": "default",
                   }],
                   "description": "test_description",
                   "identifier": {
                       "scope": "w",
                   },
                   "name": "example_id_selector",
                   "typeName": "idSelectorDefinition",
               },
               "tag": "Data",
           }],
        }
        # when undumped
        des = access.PolicyTemplateResource.model_validate(data,
            context={"style": "dump", "id": "some-id-thing"}
        )
        # the id is extracted from the context
        assert des.id == "some-id-thing"
        # the selector has discriminated to get the correct type
        assert len(des.templated_selectors) == 1
        assert des.templated_selectors[0].selector.type_name == "idSelectorDefinition"

    def test_parse_api_format(self):
        # given a response from the get api
        api = {
            "displayName": "display_name",
            "scope": "scope1",
            "code": "code1",
            "description": "some_description",
            "applications": [
                "LUSID"
            ],
            "tags": [
                "example_tag"
            ],
            "templatedSelectors": [
                {
                "application": "LUSID",
                "tag": "example_tag",
                "selector": {
                    "idSelectorDefinition": {
                        "identifier": {
                            "scope": "w"
                        },
                        "actions": [
                            {
                            "scope": "default",
                            "activity": "Read",
                            "entity": "Portfolio"
                            },
                        ],
                        "name": "example_id_selector",
                        "description": "test_description"
                        }
                    }
                }
            ]
        }
        # when parsed
        sut = access.PolicyTemplateResource.model_validate(
            api,
            context={"id": "someid"}
        )
        # then the id from the context is used for the resource id
        assert sut.id == "someid"
        # and the  selector dictionary has turned into an array
        # of selectors with a type_name on each
        assert sut.templated_selectors[0].tag == "example_tag"
        assert sut.templated_selectors[0].selector.type_name == "idSelectorDefinition"



import json
from types import SimpleNamespace

import httpx
import pytest

from fbnconfig import property, side_definition

TEST_BASE = "https://foo.lusid.com"


@pytest.mark.respx(base_url=TEST_BASE)
class DescribeSideRef:
    client = httpx.Client(base_url=TEST_BASE, event_hooks={"response": [lambda x: x.raise_for_status()]})

    def test_attach_when_present(self, respx_mock):
        # given that the remote definition exists
        respx_mock.get("/api/api/transactionconfiguration/sides/sd1?scope=sc1").mock(
            return_value=httpx.Response(200, json={})
        )
        client = self.client
        sut = side_definition.SideRef(id="one", scope="sc1", side="sd1")
        # when we call attach
        sut.attach(client)
        # then a get request was made with the scope passed as a parameter
        req = respx_mock.calls.last.request
        assert dict(req.url.params) == {"scope": "sc1"}

    def test_attach_when_missing(self, respx_mock):
        # given the remote does not exist
        respx_mock.get("/api/api/transactionconfiguration/sides/sd2?scope=sc1").mock(
            return_value=httpx.Response(404, json={})
        )
        client = self.client
        sut = side_definition.SideRef(id="two", scope="sc1", side="sd2")
        # when we call attach an exception is raised
        with pytest.raises(RuntimeError) as ex:
            sut.attach(client)
        assert "Side sd2 does not exist in scope sc1" in str(ex.value)

    def test_attach_when_http_error(self, respx_mock):
        # given the server returns an error
        respx_mock.get("/api/api/transactionconfiguration/sides/sd3?scope=sc1").mock(
            return_value=httpx.Response(400, json={})
        )
        client = self.client
        sut = side_definition.SideRef(id="three", scope="sc1", side="sd3")
        # when we call attach an exception is raised
        with pytest.raises(httpx.HTTPError):
            sut.attach(client)


@pytest.mark.respx(base_url=TEST_BASE)
class DescribeSideResource:
    client = httpx.Client(base_url=TEST_BASE, event_hooks={"response": [lambda x: x.raise_for_status()]})

    @pytest.fixture
    def property_definition_refs(self):
        names = ["a", "b", "c", "d", "e", "f"]
        return [
            property.DefinitionRef(id=name, domain=property.Domain.Transaction, scope="sc1", code=name)
            for name in names
        ]

    @pytest.fixture
    def side_resource_side_1(self):
        return side_definition.SideResource(
            id="one",
            side="sd1",
            scope="sc1",
            security="Txn:LusidInstrumentId",
            currency="Txn:TradeCurrency",
            rate="Txn:TradeToPortfolioRate",
            units="Txn:Units",
            amount="Txn:TradeAmount",
            notional_amount="0",
        )

    @pytest.fixture
    def property_definition_resources(self):
        names = ["a", "b", "c", "d", "e", "f"]
        return [
            property.DefinitionResource(
                id=name,
                domain=property.Domain.Transaction,
                scope="sc1",
                code=name,
                display_name="Example Display Name",
                data_type_id=property.ResourceId(scope="system", code="number"),
                property_description="Example property description",
                life_time=property.LifeTime.Perpetual,
                constraint_style=property.ConstraintStyle.Property,
            )
            for name in names
        ]

    @pytest.fixture
    def definition_fixtures(self, request):
        return request.getfixturevalue(request.param)

    def test_create_side(self, respx_mock, side_resource_side_1):
        respx_mock.put("/api/api/transactionconfiguration/sides/sd1?scope=sc1").mock(
            return_value=httpx.Response(200, json={})
        )
        client = self.client
        sut = side_resource_side_1
        state = sut.create(client)

        req = respx_mock.calls.last.request

        assert json.loads(req.content) == {
            "side": "sd1",
            "security": "Txn:LusidInstrumentId",
            "currency": "Txn:TradeCurrency",
            "rate": "Txn:TradeToPortfolioRate",
            "units": "Txn:Units",
            "amount": "Txn:TradeAmount",
            "notionalAmount": "0"
        }
        assert state == {"side": "sd1", "scope": "sc1"}

    @pytest.mark.parametrize(
        "definition_fixtures",
        ["property_definition_refs", "property_definition_resources"],
        indirect=True,
    )
    def test_create_side_with_property_definitions(self, respx_mock, definition_fixtures):
        prop_resources = definition_fixtures

        respx_mock.put("/api/api/transactionconfiguration/sides/sd1").mock(
            return_value=httpx.Response(200, json={})
        )
        client = self.client
        sut = side_definition.SideResource(
            id="one",
            side="sd1",
            scope="sc1",
            security=prop_resources[0],
            currency=prop_resources[1],
            rate=prop_resources[2],
            units=prop_resources[3],
            amount=prop_resources[4],
            notional_amount=prop_resources[5],
        )

        state = sut.create(client)

        req = respx_mock.calls.last.request
        assert sut.deps() == [
            prop_resources[0],
            prop_resources[1],
            prop_resources[4],
            prop_resources[2],
            prop_resources[3],
            prop_resources[5],
        ]

        assert json.loads(req.content) == {
            "side": "sd1",
            "security": "Transaction/sc1/a",
            "currency": "Transaction/sc1/b",
            "rate": "Transaction/sc1/c",
            "units": "Transaction/sc1/d",
            "amount": "Transaction/sc1/e",
            "notionalAmount": "Transaction/sc1/f",
        }
        assert state == {"side": "sd1", "scope": "sc1"}

    def test_create_side_with_mixed_definitions(
        self, respx_mock, property_definition_refs, property_definition_resources
    ):
        respx_mock.put("/api/api/transactionconfiguration/sides/sd1?scope=sc1").mock(
            return_value=httpx.Response(200, json={})
        )
        client = self.client
        sut = side_definition.SideResource(
            id="one",
            side="sd1",
            scope="sc1",
            security=property_definition_refs.pop(0),
            currency=property_definition_refs.pop(0),
            rate=property_definition_resources.pop(0),
            units=property_definition_resources.pop(0),
            amount="Txn:TradeAmount",
            notional_amount="0",
        )

        state = sut.create(client)

        req = respx_mock.calls.last.request

        assert sut.deps() == [
            property.DefinitionRef(id="a", domain=property.Domain.Transaction, scope="sc1", code="a"),
            property.DefinitionRef(id="b", domain=property.Domain.Transaction, scope="sc1", code="b"),
            property.DefinitionResource(
                id="a",
                domain=property.Domain.Transaction,
                scope="sc1",
                code="a",
                display_name="Example Display Name",
                data_type_id=property.ResourceId(scope="system", code="number"),
                property_description="Example property description",
                life_time=property.LifeTime.Perpetual,
                constraint_style=property.ConstraintStyle.Property,
            ),
            property.DefinitionResource(
                id="b",
                domain=property.Domain.Transaction,
                scope="sc1",
                code="b",
                display_name="Example Display Name",
                data_type_id=property.ResourceId(scope="system", code="number"),
                property_description="Example property description",
                life_time=property.LifeTime.Perpetual,
                constraint_style=property.ConstraintStyle.Property,
            ),
        ]

        assert json.loads(req.content) == {
            "side": "sd1",
            "security": "Transaction/sc1/a",
            "currency": "Transaction/sc1/b",
            "rate": "Transaction/sc1/a",
            "units": "Transaction/sc1/b",
            "amount": "Txn:TradeAmount",
            "notionalAmount": "0",
        }
        assert state == {"side": "sd1", "scope": "sc1"}

    def test_read_side(self, respx_mock, side_resource_side_1):
        respx_mock.get("/api/api/transactionconfiguration/sides/sd1?scope=sc1").mock(
            return_value=httpx.Response(200, json=side_resource_side_1.model_dump(by_alias=True))
        )

        client = self.client
        old_state = SimpleNamespace(scope="sc1", side="sd1")

        sut = side_resource_side_1

        remote = sut.read(client, old_state)
        assert remote == side_resource_side_1.model_dump(by_alias=True)
        assert {"scope": old_state.scope, "side": old_state.side} == {
            "scope": sut.scope,
            "side": sut.side,
        }

    def test_read_side_failure_not_found(self, respx_mock, side_resource_side_1):
        respx_mock.get("/api/api/transactionconfiguration/sides/sd1?scope=sc1").mock(
            return_value=httpx.Response(404)
        )
        old_state = SimpleNamespace(scope="sc1", side="sd1")
        client = self.client
        sut = side_resource_side_1

        with pytest.raises(httpx.HTTPStatusError):
            remote = sut.read(client, old_state)
            assert remote is None

    def test_delete_side(self, respx_mock, side_resource_side_1):
        respx_mock.delete("/api/api/transactionconfiguration/sides/sd1/$delete?scope=sc1").mock(
            return_value=httpx.Response(200)
        )
        client = self.client
        old_state = SimpleNamespace(scope="sc1", side="sd1")
        sut = side_resource_side_1

        sut.delete(client, old_state)

    def test_delete_side_missing(self, respx_mock, side_resource_side_1):
        respx_mock.delete("/api/api/transactionconfiguration/sides/sd1/$delete?scope=sc1").mock(
            return_value=httpx.Response(404)
        )
        client = self.client
        old_state = SimpleNamespace(scope="sc1", side="sd1")
        sut = side_resource_side_1

        with pytest.raises(httpx.HTTPStatusError):
            sut.delete(client, old_state)

    def test_update_side_no_change(self, respx_mock, side_resource_side_1):
        respx_mock.get("/api/api/transactionconfiguration/sides/sd1?scope=sc1").mock(
            return_value=httpx.Response(200, json=side_resource_side_1.model_dump(by_alias=True))
        )

        client = self.client

        sut = side_resource_side_1
        old_state = SimpleNamespace(scope="sc1", side="sd1")
        state = sut.update(client, old_state)
        assert state is None

    @pytest.mark.parametrize(
        "update_changes",
        [
            {
                "does_require_get": False,
                "put_side": "sd1",
                "put_scope": "sc2",
                "put_response": {
                    "side": "sd1",
                    "security": "Txn:LusidInstrumentId",
                    "currency": "Txn:TradeCurrency",
                    "rate": "Txn:TradeToPortfolioRate",
                    "units": "Txn:Units",
                    "amount": "Txn:TradeAmount",
                    "notionalAmount": "0",
                },
                "resource": side_definition.SideResource(
                    id="two",
                    side="sd1",
                    scope="sc2",
                    security="Txn:LusidInstrumentId",
                    currency="Txn:TradeCurrency",
                    rate="Txn:TradeToPortfolioRate",
                    units="Txn:Units",
                    amount="Txn:TradeAmount",
                    notional_amount="0",
                ),
            },
            {
                "does_require_get": False,
                "put_side": "sd2",
                "put_scope": "sc1",
                "put_response": {
                    "side": "sd2",
                    "security": "Txn:LusidInstrumentId",
                    "currency": "Txn:TradeCurrency",
                    "rate": "Txn:TradeToPortfolioRate",
                    "units": "Txn:Units",
                    "amount": "Txn:TradeAmount",
                    "notionalAmount": "0",
                },
                "resource": side_definition.SideResource(
                    id="two",
                    side="sd2",
                    scope="sc1",
                    security="Txn:LusidInstrumentId",
                    currency="Txn:TradeCurrency",
                    rate="Txn:TradeToPortfolioRate",
                    units="Txn:Units",
                    amount="Txn:TradeAmount",
                    notional_amount="0",
                ),
            },
            {
                "does_require_get": True,
                "put_side": "sd1",
                "put_scope": "sc1",
                "put_response": {
                    "side": "sd1",
                    "security": "Txn:TradeCurrency",
                    "currency": "Txn:TradeCurrency",
                    "rate": "Txn:TradeToPortfolioRate",
                    "units": "Txn:Units",
                    "amount": "Txn:TradeAmount",
                    "notionalAmount": "0",
                },
                "resource": side_definition.SideResource(
                    id="one",
                    side="sd1",
                    scope="sc1",
                    security="Txn:TradeCurrency",
                    currency="Txn:TradeCurrency",
                    rate="Txn:TradeToPortfolioRate",
                    units="Txn:Units",
                    amount="Txn:TradeAmount",
                    notional_amount="0",
                ),
                "get_side": "sd1",
                "get_scope": "sc1",
                "get_response": {
                    "side": "sd1",
                    "security": "Txn:LusidInstrumentId",
                    "currency": "Txn:TradeCurrency",
                    "rate": "Txn:TradeToPortfolioRate",
                    "units": "Txn:Units",
                    "amount": "Txn:TradeAmount",
                    "notionalAmount": "0",
                },
            },
        ],
    )
    def test_update_side(self, respx_mock, update_changes):
        respx_mock.put(
            f"/api/api/transactionconfiguration/sides/{update_changes['put_side']}?scope={update_changes['put_scope']}"
        ).mock(return_value=httpx.Response(200, json=update_changes["put_response"]))

        if update_changes["does_require_get"]:
            respx_mock.get(
                f"/api/api/transactionconfiguration/sides/{update_changes['get_side']}?scope={update_changes['get_scope']}"
            ).mock(return_value=httpx.Response(200, json=update_changes["get_response"]))
        else:
            respx_mock.delete("/api/api/transactionconfiguration/sides/sd1/$delete?scope=sc1").mock(
                return_value=httpx.Response(200)
            )

        client = self.client
        old_state = SimpleNamespace(scope="sc1", side="sd1")
        sut = update_changes["resource"]
        state = sut.update(client, old_state)

        if update_changes["does_require_get"]:
            req = respx_mock.calls.last.request
            assert state == {"scope": old_state.scope, "side": old_state.side}
            assert json.loads(req.content) == update_changes["put_response"]
        else:
            assert state != old_state

    def test_deps(self, property_definition_refs):
        sut = side_definition.SideResource(
            id="one",
            side="sd1",
            scope="sc1",
            security=property_definition_refs[0],
            currency=property_definition_refs[1],
            rate=property_definition_refs[2],
            units=property_definition_refs[2],
            amount="Txn:TradeAmount",
            notional_amount="0",
        )
        deps = sut.deps()
        assert deps == [
            property.DefinitionRef(id="a", domain=property.Domain.Transaction, scope="sc1", code="a"),
            property.DefinitionRef(id="b", domain=property.Domain.Transaction, scope="sc1", code="b"),
            property.DefinitionRef(id="c", domain=property.Domain.Transaction, scope="sc1", code="c"),
            property.DefinitionRef(id="c", domain=property.Domain.Transaction, scope="sc1", code="c"),
        ]

    def test_dump(self, side_resource_side_1):
        # given a side resource with string values (no property refs)
        sut = side_resource_side_1
        # when we dump it
        result = sut.model_dump(
            mode="json",
            by_alias=True,
            round_trip=True,
            exclude_none=True,
            context={"style": "dump"}
        )
        # then it includes scope and side information with string values
        expected = {
            "side": "sd1",
            "scope": "sc1",
            "security": "Txn:LusidInstrumentId",
            "currency": "Txn:TradeCurrency",
            "rate": "Txn:TradeToPortfolioRate",
            "units": "Txn:Units",
            "amount": "Txn:TradeAmount",
            "notionalAmount": "0"
        }
        assert result == expected

    def test_dump_with_property_refs(self, property_definition_refs):
        # given a side resource with property definition references
        sut = side_definition.SideResource(
            id="side_with_refs",
            side="sd1",
            scope="sc1",
            security=property_definition_refs[0],  # DefinitionRef
            currency=property_definition_refs[1],  # DefinitionRef
            rate="Txn:TradeToPortfolioRate",       # string
            units="Txn:Units",                     # string
            amount=property_definition_refs[2],    # DefinitionRef
            notional_amount="0"                    # string
        )
        # when we dump it
        result = sut.model_dump(
            mode="json",
            by_alias=True,
            round_trip=True,
            exclude_none=True,
            context={"style": "dump"}
        )
        # then property refs are serialized as $ref and strings remain as strings
        expected = {
            "side": "sd1",
            "scope": "sc1",
            "security": {"$ref": "a"},
            "currency": {"$ref": "b"},
            "rate": "Txn:TradeToPortfolioRate",
            "units": "Txn:Units",
            "amount": {"$ref": "c"},
            "notionalAmount": "0"
        }
        assert result == expected

    def test_undump(self):
        # given dump data with string values
        data = {
            "side": "sd1",
            "scope": "sc1",
            "security": "Txn:LusidInstrumentId",
            "currency": "Txn:TradeCurrency",
            "rate": "Txn:TradeToPortfolioRate",
            "units": "Txn:Units",
            "amount": "Txn:TradeAmount",
            "notionalAmount": "0"
        }
        # when we undump it
        result = side_definition.SideResource.model_validate(
            data, context={"style": "undump", "id": "side_id"}
        )
        # then it's correctly populated
        assert result.id == "side_id"
        assert result.side == "sd1"
        assert result.scope == "sc1"
        assert result.security == "Txn:LusidInstrumentId"
        assert result.currency == "Txn:TradeCurrency"
        assert result.rate == "Txn:TradeToPortfolioRate"
        assert result.units == "Txn:Units"
        assert result.amount == "Txn:TradeAmount"
        assert result.notional_amount == "0"

    def test_undump_with_property_refs(self, property_definition_refs):
        # given dump data with $ref values
        data = {
            "side": "sd1",
            "scope": "sc1",
            "security": {"$ref": "a"},
            "currency": {"$ref": "b"},
            "rate": "Txn:TradeToPortfolioRate",
            "units": "Txn:Units",
            "amount": {"$ref": "c"},
            "notionalAmount": "0"
        }
        # when we undump it with $refs context
        refs_dict = {ref.id: ref for ref in property_definition_refs}
        result = side_definition.SideResource.model_validate(
            data, context={"style": "undump", "$refs": refs_dict, "id": "side_id"}
        )
        # then property refs are resolved and strings remain as strings
        assert result.id == "side_id"
        assert result.side == "sd1"
        assert result.scope == "sc1"
        assert result.security == property_definition_refs[0]  # resolved DefinitionRef
        assert result.currency == property_definition_refs[1]  # resolved DefinitionRef
        assert result.rate == "Txn:TradeToPortfolioRate"       # string
        assert result.units == "Txn:Units"                     # string
        assert result.amount == property_definition_refs[2]    # resolved DefinitionRef
        assert result.notional_amount == "0"                   # string



import os
from types import SimpleNamespace

from pytest import fixture

import fbnconfig
from tests.integration import workflows
from tests.integration.generate_test_name import gen

if os.getenv("FBN_ACCESS_TOKEN") is None or os.getenv("LUSID_ENV") is None:
    raise (
        RuntimeError("Both FBN_ACCESS_TOKEN and LUSID_ENV variables need to be set to run these tests")
    )

lusid_env = os.environ["LUSID_ENV"]
token = os.environ["FBN_ACCESS_TOKEN"]
host_vars = {}
client = fbnconfig.create_client(lusid_env, token)


@fixture
def setup_deployment():
    deployment_name = gen("workflows")
    print(f"\nRunning for deployment {deployment_name}...")
    yield SimpleNamespace(name=deployment_name)
    # Teardown: Clean up resources (if any) after the test
    print("\nTearing down resources...")
    fbnconfig.deploy(fbnconfig.Deployment(deployment_name, []), lusid_env, token)


def test_teardown(setup_deployment):
    # create first
    fbnconfig.deploy(workflows.configure(setup_deployment), lusid_env, token)
    fbnconfig.deploy(fbnconfig.Deployment(setup_deployment.name, []), lusid_env, token)
    result = client.get(f"/workflow/api/workers?filter=id.scope eq '{setup_deployment.name}'")
    # no response returned
    assert len(result.json()["values"]) == 0


def test_create(setup_deployment):
    fbnconfig.deploy(workflows.configure(setup_deployment), lusid_env, token)
    workers = client.get(f"/workflow/api/workers?filter=id.scope eq '{setup_deployment.name}'")
    task_defs = client.get(f"/workflow/api/taskdefinitions?filter=id.scope eq '{setup_deployment.name}'")
    event_handlers = client.get(
        f"/workflow/api/eventhandlers?filter=id.scope eq '{setup_deployment.name}'"
    )
    assert len(workers.json()["values"]) == 1
    assert len(task_defs.json()["values"]) == 1
    assert len(event_handlers.json()["values"]) == 2


def test_update_nochange(setup_deployment):
    fbnconfig.deploy(workflows.configure(setup_deployment), lusid_env, token)
    actions = fbnconfig.deploy(workflows.configure(setup_deployment), lusid_env, token)
    assert {a.change for a in actions} == {"nochange", "attach"}
    workers = client.get(f"/workflow/api/workers?filter=id.scope eq '{setup_deployment.name}'")
    task_defs = client.get(f"/workflow/api/taskdefinitions?filter=id.scope eq '{setup_deployment.name}'")
    event_handlers = client.get(
        f"/workflow/api/eventhandlers?filter=id.scope eq '{setup_deployment.name}'"
    )
    assert len(workers.json()["values"]) == 1
    assert len(task_defs.json()["values"]) == 1
    assert len(event_handlers.json()["values"]) == 2
